import numpy as np
import numpy.testing
import pytest
from bluesky import plans as bp
from bluesky.run_engine import RunEngine
from ophyd.positioner import LimitError

from .. import SimulatedE4CV
from ..calc import UnreachableError


class Fourc(SimulatedE4CV):
    """4-circle for testing."""


def check_hkl(diffractometer, h, k, l):
    try:
        diffractometer.check_value({"h": h, "k": k, "l": l})
    except UnreachableError as exc:
        assert False, f"({h}, {k}, {l}) : {exc}"


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    return fourc


def test_initial(fourc):
    assert fourc.h.position == 0
    assert fourc.k.position == 0
    assert fourc.l.position == 0


def test_limits(fourc):
    assert fourc.h.limits == (0, 0)
    assert fourc.h.check_value(1) is None
    assert fourc.h.check_value(1.2) is None
    with pytest.raises(ValueError) as exinfo:
        fourc.h.check_value(10)
    assert "Unable to solve." in str(exinfo.value)


def test_check_value(fourc):
    assert fourc.check_value((1, 0, 0)) is None
    assert fourc.check_value(dict(h=1, k=0, l=0)) is None
    assert fourc.check_value(dict(h=1, k=0)) is None

    with pytest.raises(TypeError) as exinfo:
        assert fourc.check_value(1, 0, 0) is None
    assert "check_value() takes 2 positional arguments" in str(exinfo.value)

    with pytest.raises(ValueError) as exinfo:
        assert fourc.check_value(1) is None
    assert "Not all required values for a PseudoPosition" in str(exinfo.value)

    assert fourc.check_value(dict(h=1, tth=0)) is None

    with pytest.raises(LimitError) as exinfo:
        fourc.check_value(dict(h=1, tth=np.inf))
    assert "position=inf not within limits" in str(exinfo.value)

    with pytest.raises(KeyError) as exinfo:
        fourc.check_value(dict(waldo=0))
    assert "waldo not in fourc" in str(exinfo.value)


def test_move(fourc):
    ppos = (1.2, 1.2, 0.001)
    rpos = (
        58.051956519652485,
        44.99999005281953,
        89.95225352812487,
        116.10391303930497,
    )
    fourc.move(ppos)
    numpy.testing.assert_almost_equal(fourc.position, ppos)
    numpy.testing.assert_almost_equal(tuple(fourc.real_position), rpos)


@pytest.mark.parametrize("start", [[1.2, 1.2, 0.001], [1, 0, 0], [1, 1, 1]])
@pytest.mark.parametrize("h", np.arange(0.9, 1.1, 0.1))
@pytest.mark.parametrize("k", np.arange(0.0, 1.2, 0.6))
@pytest.mark.parametrize("l", np.arange(0, 1, 0.5))
def test_hkl_scan(start, h, k, l, fourc):
    assert len(start) == 3
    fourc.move(start)
    check_hkl(fourc, h, k, l)


def test_hkl_range_error(fourc):
    with pytest.raises(UnreachableError) as exinfo:
        fourc.check_value({"h": 0.9, "k": 0.9, "l": 123})
    assert "Unable to solve." in str(exinfo.value)


@pytest.mark.parametrize("start", [[1, 1, 1]])
@pytest.mark.parametrize("tth", np.arange(10, 20, 4))
def test_real_axis(start, tth, fourc):
    assert len(start) == 3
    fourc.move(start)
    try:
        fourc.check_value({"tth": tth})
    except UnreachableError as exc:
        assert False, f"{exc}"


@pytest.mark.parametrize("start", [[1, 1, 1]])
@pytest.mark.parametrize(
    "target",
    [
        {"tth": 10},
        {"tth": 20},
        {"tth": 20, "chi": 7},
    ],
)
def test_moves(start, target, fourc):
    assert len(start) == 3
    assert isinstance(target, dict)
    fourc.move(start)
    try:
        fourc.inverse(target)
    except UnreachableError as exc:
        assert False, f"{target=} : {exc}"


def test_axis_contention(fourc):
    RE = RunEngine()
    # contention if move pseudo and real positioners together
    with pytest.raises(ValueError) as exinfo:
        RE(bp.scan([fourc], fourc.tth, 10, 20, fourc.k, 0, 0, 3))
    assert "mix of real and pseudo" in str(exinfo.value)


def test_real_axis_range_error(fourc):
    with pytest.raises(LimitError) as exinfo:
        fourc.check_value({"tth": 10_000})
    assert "not within limits" in str(exinfo.value)



"""
Test that reflections from one sample are not restored to other samples.
"""

from .. import DiffractometerConfiguration
from .tools import sample_kryptonite
from .tools import sample_silicon
from .tools import sample_vibranium


def test_issue289(e4cv):
    assert e4cv is not None

    # setup four samples with reflections, all of them different
    main = e4cv.calc.sample  # the default sample
    assert len(main.reflections) == 0
    m_100 = main.add_reflection(1, 0, 0, (-45, 0, 0, 0))
    m_010 = main.add_reflection(0, 1, 0, (45, 0, 0, 0))
    main.compute_UB(m_100, m_010)

    kryptonite = sample_kryptonite(e4cv)
    assert len(kryptonite.reflections) == 0
    k_200 = kryptonite.add_reflection(2, 0, 0, (30, 0, 10, 60))
    k_020 = kryptonite.add_reflection(0, 2, 0, (30, 90, 10, 60))
    kryptonite.compute_UB(k_200, k_020)

    vibranium = sample_vibranium(e4cv)
    assert len(vibranium.reflections) == 0
    v_003 = vibranium.add_reflection(0, 0, 3, (20.33, 4.33, 8.33, 40.33))
    v_033 = vibranium.add_reflection(0, 3, 3, (20.33, -94.33, 8.33, 40.33))
    vibranium.compute_UB(v_003, v_033)

    silicon = sample_silicon(e4cv)
    assert len(silicon.reflections) == 0
    s_440 = silicon.add_reflection(4, 4, 0, (34, 44, 54, 64))
    s_444 = silicon.add_reflection(4, 4, 4, (34, 134, 54, 64))
    silicon.compute_UB(s_440, s_444)

    n_saved_reflections = 2
    assert len(main.reflections) == n_saved_reflections
    assert len(kryptonite.reflections) == n_saved_reflections
    assert len(silicon.reflections) == n_saved_reflections
    assert len(vibranium.reflections) == n_saved_reflections
    # same test, using diffractometer sample dictionary now.
    for sample in e4cv.calc._samples.values():
        assert len(sample.reflections) == n_saved_reflections, f"{sample.name=}"
    assert len(e4cv.calc._samples) == 4

    agent = DiffractometerConfiguration(e4cv)
    assert agent is not None
    config = agent.export()
    assert isinstance(config, str)

    # Restore the configuration without clearing the diffractometer first.
    # This should not change the number of samples or the number of
    # reflections for any of the samples.
    agent.restore(config, clear=False)

    assert len(e4cv.calc._samples) == 4
    for sample in e4cv.calc._samples.values():
        assert len(sample.reflections) == 2 * n_saved_reflections



import time

import bluesky.plans as bp
import databroker
import numpy.testing
import pandas as pd
import pytest
from bluesky import RunEngine
from bluesky import plan_stubs as bps
from ophyd.sim import hw

import hkl.util
from hkl import SimulatedE4CV
from hkl import SimulatedK4CV
from hkl.calc import A_KEV

from .tools import sample_silicon


class Fourc(SimulatedE4CV):
    """4-circle for testing."""


class Kappa(SimulatedK4CV):
    """Kappa 4-circle for testing."""


@pytest.fixture
def cat():
    return databroker.temp().v2


@pytest.fixture
def RE(cat):
    engine = RunEngine({})
    engine.subscribe(cat.v1.insert)
    return engine


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    fourc.wait_for_connection()
    fourc._update_calc_energy()

    fourc.energy.put(A_KEV / 1.54)
    sample = sample_silicon(fourc)
    r_400 = sample.add_reflection(4, 0, 0, (-145.451, 0, 0, 69.0966))
    r_040 = sample.add_reflection(0, 4, 0, (-145.451, 0, 90, 69.0966))
    sample.compute_UB(r_400, r_040)

    return fourc


@pytest.fixture(scope="function")
def kappa():
    kappa = Kappa("", name="kappa")
    kappa.wait_for_connection()
    kappa._update_calc_energy()

    kappa.energy.put(A_KEV / 1.54)
    sample = sample_silicon(kappa)
    r_400 = sample.add_reflection(4, 0, 0, (55.4507, 0, 90, -69.0966))
    r_040 = sample.add_reflection(0, 4, 0, (-1.5950, 134.7568, 123.3554, -69.0966))

    sample.compute_UB(r_400, r_040)

    return kappa


def test_basic_setup(fourc, kappa):
    assert fourc is not None
    assert kappa is not None
    assert fourc != kappa
    assert fourc._reals != kappa._reals

    assert round(fourc.calc.wavelength, 2) == 1.54
    assert fourc.calc.wavelength == kappa.calc.wavelength


def test_fourc_orientation_save(cat, RE, fourc):
    assert len(cat) == 0
    det = hw().noisy_det

    # this run will not save orientation information
    _uids = RE(bp.count([det]))
    assert len(_uids) == 1
    time.sleep(1)
    assert len(cat) == 1
    assert "fourc" not in cat[1].primary.config

    # this run _will_ save orientation information
    _uids = RE(bp.count([det, fourc]))
    assert len(_uids) == 1
    time.sleep(1)
    assert len(cat) == 2

    key_list = """
        _pseudos
        _reals
        class_name
        diffractometer_name
        geometry_name
        lattice
        orientation_attrs
        reflections_details
        sample_name
        UB
    """.split()

    if databroker.__version__ < "2.0":
        conf = cat[_uids[0]].primary.config["fourc"].read()
        assert not isinstance(conf, dict)
        descriptors = conf.to_dict()
        assert isinstance(descriptors, dict)
        assert list(descriptors.keys()) == "coords attrs dims data_vars".split()
        for key in key_list:
            key_name = f"fourc_{key}"
            assert hasattr(conf, key_name), f"{key_name=!r} {list(conf)=!r}"
            assert key_name in descriptors["data_vars"]

        assert conf.fourc_class_name == "Fourc"
        assert conf.fourc_geometry_name == "E4CV"
        assert conf.fourc_diffractometer_name == "fourc"
        assert conf.fourc_sample_name == "silicon"

        assert len(conf.fourc__pseudos) == 1
        assert conf.fourc__pseudos[0].values.tolist() == "h k l".split()
        assert len(conf.fourc__reals) == 1
        assert conf.fourc__reals.values[0].tolist() == "omega chi phi tth".split()

    else:
        descriptors = cat[_uids[0]].primary.descriptors
        assert isinstance(descriptors, list)
        assert len(descriptors) > 0
        for descriptor in descriptors:
            for device, configuration in descriptor.get("configuration", {}).items():
                assert isinstance(device, str)
                assert isinstance(configuration, dict)
                conf = configuration.get("data", {})
                assert isinstance(conf, dict)

                if device != "fourc":
                    continue
                for key in key_list:
                    key_name = f"fourc_{key}"
                    assert key_name in conf

                assert conf["fourc_class_name"] == "Fourc"
                assert conf["fourc_geometry_name"] == "E4CV"
                assert conf["fourc_diffractometer_name"] == "fourc"
                assert conf["fourc_sample_name"] == "silicon"

                assert len(conf["fourc__pseudos"]) == 3
                assert conf["fourc__pseudos"] == "h k l".split()
                assert len(conf["fourc__reals"]) == 4
                assert conf["fourc__reals"] == "omega chi phi tth".split()


def test_fourc_run_orientation_info(cat, RE, fourc):
    RE(bp.count([fourc]))
    info = hkl.util.run_orientation_info(cat[1])
    assert info is not None
    assert isinstance(info, dict)
    assert len(info) > 0, f"{info=!r}"
    assert "fourc" in info
    fourc_orient = info["fourc"]
    assert "orientation_attrs" in fourc_orient
    assert "reflections_details" in fourc_orient["orientation_attrs"]
    assert "reflections_details" in fourc_orient
    refls = fourc_orient["reflections_details"]
    assert len(refls) == 2
    assert "wavelength" in refls[0]
    assert round(refls[0]["wavelength"], 2) == 1.54


def test_list_orientation_runs(cat, RE, fourc, kappa):
    det = hw().noisy_det

    def scans():
        yield from bp.count([det])
        yield from bp.count([fourc])
        yield from bp.count([kappa])
        yield from bp.count([fourc, kappa])

    RE(scans())
    runs = hkl.util.list_orientation_runs(cat)
    assert len(runs) == 4, f"{runs=!r}"
    # four sets of orientation info
    # (last scan has 2, first scan has none)
    assert len(runs.scan_id) == 4
    assert 1 not in runs.scan_id.to_list(), f"{runs=!r}"  # no orientation
    assert runs.scan_id.to_list() == [2, 3, 4, 4]
    assert runs.diffractometer_name.to_list() == "fourc kappa fourc kappa".split()


def test_no_primary_stream(cat, RE, fourc):
    det = hw().noisy_det

    def my_plan():
        yield from bps.open_run()
        yield from bps.create(name="not_primary")
        yield from bps.read(det)
        yield from bps.save()
        yield from bps.close_run()

    def scans():
        yield from bp.count([fourc])
        yield from my_plan()

    uids = RE(scans())
    assert len(uids) == 2

    runs = hkl.util.list_orientation_runs(cat)
    # my_plan() has no primary stream
    assert len(runs.scan_id) == 1


def test_missing_energy_key(cat, RE, fourc):
    """Issue 216."""

    def scans():
        yield from bp.count([fourc])

    uids = RE(scans())
    assert len(uids) == 1, f"{uids=!r}"
    assert uids[0] in cat

    runs = hkl.util.list_orientation_runs(cat)
    assert isinstance(runs, pd.DataFrame)

    run = cat[uids[0]]
    orientations = hkl.util.run_orientation_info(run)
    assert len(orientations) == 1

    assert fourc.name in orientations
    orientation = orientations[fourc.name]
    assert "energy" in orientation

    # trigger the error by removing the "energy" key
    with pytest.raises(KeyError) as exinfo:
        orientation.pop("energy")
        hkl.util.restore_energy(orientation, fourc)
    assert " Cannot restore diffractometer energy " in str(exinfo.value)


def test_restore_orientation(cat, RE, fourc):
    RE(bp.count([fourc]))
    info = hkl.util.run_orientation_info(cat[-1])
    assert isinstance(info, dict)

    fourc_orient = info.get("fourc")
    assert fourc_orient is not None, f"{fourc_orient=!r}"

    e4cv = Fourc("", name="e4cv")
    assert len(e4cv.calc._samples) == 1

    # typical case : good match, restores successfully
    hkl.util.restore_orientation(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    sample = e4cv.calc.sample
    assert sample.name == "silicon"
    assert len(sample.reflections) == 2
    numpy.testing.assert_array_equal(sample.reflections, [[4, 0, 0], [0, 4, 0]])
    # fmt: off
    numpy.testing.assert_array_almost_equal(
        fourc.calc.sample.UB,
        e4cv.calc.sample.UB,
    )
    # fmt: on

    # geometry mismatch, cannot restore
    k4cv = Kappa("", name="k4cv")
    assert len(k4cv.calc._samples) == 1
    assert len(k4cv.calc.sample.reflections) == 0
    with pytest.raises(ValueError) as exinfo:
        hkl.util.restore_orientation(fourc_orient, k4cv)
    args = exinfo.value.args
    assert len(args) == 1
    expected = "Geometries do not match: Orientation=E4CV, k4cv=K4CV, will not restore."
    assert args[0] == expected

    # different class name, restores successfully
    e4cv = SimulatedE4CV("", name="e4cv")
    assert len(e4cv.calc._samples) == 1
    assert len(e4cv.calc.sample.reflections) == 0
    hkl.util.restore_orientation(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    assert len(e4cv.calc.sample.reflections) == 2
    # fmt: off
    numpy.testing.assert_array_almost_equal(
        fourc.calc.sample.UB,
        e4cv.calc.sample.UB,
    )
    # fmt: on

    # different real axis names, restores successfully
    e4cv = Fourc("", name="e4cv")
    # change the real axis names
    e4cv.calc.physical_axis_names = {
        "omega": "john",
        "chi": "paul",
        "phi": "george",
        "tth": "ringo",
    }
    assert len(e4cv.calc._samples) == 1
    assert len(e4cv.calc.sample.reflections) == 0
    hkl.util.restore_orientation(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    assert len(e4cv.calc.sample.reflections) == 2
    # fmt: off
    numpy.testing.assert_array_almost_equal(
        fourc.calc.sample.UB,
        e4cv.calc.sample.UB,
    )
    # fmt: on


def test_restore_sample(cat, RE, fourc):
    RE(bp.count([fourc]))
    info = hkl.util.run_orientation_info(cat[-1])
    assert isinstance(info, dict)

    fourc_orient = info.get("fourc")
    assert fourc_orient is not None, f"{fourc_orient=!r}"

    # Add sample to new diffractometer.
    e4cv = Fourc("", name="e4cv")
    hkl.util.restore_sample(fourc_orient, e4cv)
    assert len(e4cv.calc._samples) == 2
    assert len(e4cv.calc.sample.reflections) == 0
    assert e4cv.calc.sample.name == fourc.calc.sample.name

    # Sample already defined, will not make another.
    with pytest.raises(ValueError) as exinfo:
        hkl.util.restore_sample(fourc_orient, e4cv)
    expected = "Sample 'silicon' already exists in e4cv."
    assert exinfo.value.args[0] == expected



"""
Common structures for testing.
"""

import pytest
from ophyd import Component
from ophyd import SoftPositioner

from .. import E4CV
from .. import E6C
from .. import SimMixin
from .. import SimulatedE4CV
from .. import SimulatedE6C
from .. import SimulatedK4CV
from .. import SimulatedK6C


@pytest.fixture
def e4cv():
    """Standard E4CV."""
    diffractometer = SimulatedE4CV("", name="e4cv")
    diffractometer.wait_for_connection()
    diffractometer._update_calc_energy()
    return diffractometer


@pytest.fixture
def e4cv_renamed():
    """E4CV with renamed axes."""

    class CustomFourCircle(SimMixin, E4CV):
        """E4CV with renamed axes."""

        theta = Component(SoftPositioner, kind="hinted", init_pos=0)
        chi = Component(SoftPositioner, kind="hinted", init_pos=0)
        phi = Component(SoftPositioner, kind="hinted", init_pos=0)
        ttheta = Component(SoftPositioner, kind="hinted", init_pos=0)

    diffractometer = CustomFourCircle("", name="e4cv_renamed")
    # rename the physical axes
    diffractometer.calc.physical_axis_names = {
        # E4CV: local
        "omega": "theta",
        "chi": "chi",
        "phi": "phi",
        "tth": "ttheta",
    }
    return diffractometer


@pytest.fixture
def e6c():
    """Standard E6C."""
    diffractometer = SimulatedE6C("", name="e6cv")
    diffractometer.wait_for_connection()
    diffractometer._update_calc_energy()
    return diffractometer


@pytest.fixture
def k4cv():
    """Standard K4CV."""
    diffractometer = SimulatedK4CV("", name="k4cv")
    diffractometer.wait_for_connection()
    diffractometer._update_calc_energy()
    return diffractometer


@pytest.fixture
def k6c():
    """Standard K6C."""
    diffractometer = SimulatedK6C("", name="k6c")
    diffractometer.wait_for_connection()
    diffractometer._update_calc_energy()
    return diffractometer


@pytest.fixture
def tardis():
    class Tardis(SimMixin, E6C):
        """E6C variant at NSLS-II."""

        # theta
        theta = Component(SoftPositioner, init_pos=0)
        omega = Component(SoftPositioner, init_pos=0)
        chi = Component(SoftPositioner, init_pos=0)
        phi = Component(SoftPositioner, init_pos=0)
        # delta, gamma
        delta = Component(SoftPositioner, init_pos=0)
        gamma = Component(SoftPositioner, init_pos=0)

    diffractometer = Tardis("", name="tardis")
    diffractometer.calc.engine.mode = "lifting_detector_mu"
    # re-map Tardis' axis names onto what an E6C expects
    diffractometer.calc.physical_axis_names = {
        "mu": "theta",
        "omega": "omega",
        "chi": "chi",
        "phi": "phi",
        "gamma": "delta",
        "delta": "gamma",
    }
    diffractometer.wait_for_connection()
    return diffractometer



"""
Common code, setups, constants, ... for these tests.

Avoids direct imports of __init__.py.
"""

import json
import logging
from pathlib import Path  # noqa

import numpy

from ..util import new_lattice

logger = logging.getLogger("ophyd_session_test")

TARDIS_TEST_MODE = "lifting_detector_mu"
TWO_PI = 2 * numpy.pi


class DocsCollector:
    """Collect representative documents from the RE."""

    def __init__(self) -> None:
        self.raw = {}
        self.safe_json = {}

    def receiver(self, key, doc):
        self.raw[key] = doc
        self.safe_json[key] = safe_json_dump(doc)

    def to_file(self, fpath: (Path, str), indent: int = 2) -> None:  # type: ignore
        """Write collected docs to a file."""
        with open(fpath, "w") as f:
            dd = {k: json.loads(v.decode("utf8")) for k, v in self.safe_json.items()}
            json.dump(dd, f, indent=indent)


def new_sample(diffractometer, name, lattice):
    return diffractometer.calc.new_sample(name, lattice=lattice)


def sample_kryptonite(diffractometer):
    triclinic = new_lattice(4, 5, 6, 75, 85, 95)
    return new_sample(diffractometer, "kryptonite", lattice=triclinic)


def sample_silicon(diffractometer):
    from .. import SI_LATTICE_PARAMETER

    cubic = new_lattice(SI_LATTICE_PARAMETER)
    return new_sample(diffractometer, "silicon", lattice=cubic)


def sample_vibranium(diffractometer):
    a0 = TWO_PI
    cubic = new_lattice(a0)
    return new_sample(diffractometer, "vibranium", lattice=cubic)


def validate_descriptor_doc_content(gname, descriptor):
    """Validate the descriptor document content."""
    assert isinstance(descriptor, dict)

    config = descriptor.get("configuration")
    assert isinstance(config, dict)

    diffractometer_config = config.get(gname)
    assert isinstance(diffractometer_config, dict)

    assert "data" in diffractometer_config
    data = diffractometer_config.get("data")
    assert isinstance(data, dict)

    attrs = [
        ["sample_name", str],
        ["lattice", (numpy.ndarray, list)],
        ["lattice_reciprocal", (tuple, list)],
        ["U", (numpy.ndarray, list)],
        ["UB", (numpy.ndarray, list)],
        ["reflections_details", list],
        ["_pseudos", (tuple, list)],
        ["_reals", (tuple, list)],
        ["_constraints", (numpy.ndarray, list)],
        ["ux", float],
        ["uy", float],
        ["uz", float],
        ["orientation_attrs", list],
    ]
    for attr, struct in attrs:
        assert isinstance(data.get(f"{gname}_{attr}"), struct), f"{attr=!r} {list(data)=!r}"


def safe_json_dump(content):
    """
    Vendored from tiled.utils

    https://github.com/bluesky/tiled/blob/15abbc98df9bfd25f8c713656f664a558136e26f/tiled/utils.py#L526

    Base64-encode raw bytes, and provide a fallback if orjson numpy handling fails.
    """
    import base64
    import sys

    import orjson

    def default(content):
        if isinstance(content, bytes):
            content = f"data:application/octet-stream;base64,{base64.b64encode(content).decode('utf-8')}"
            return content
        if isinstance(content, Path):
            return str(content)
        # No need to import numpy if it hasn't been used already.
        numpy = sys.modules.get("numpy", None)
        if numpy is not None:
            if isinstance(content, numpy.ndarray):
                # If we make it here, OPT_NUMPY_SERIALIZE failed because we have hit some edge case.
                # Give up on the numpy fast-path and convert to Python list.
                # If the items in this list aren't serializable (e.g. bytes) we'll recurse on each item.
                return content.tolist()
            elif isinstance(content, (bytes, numpy.bytes_)):
                return content.decode("utf-8")
        raise TypeError

    # Not all numpy dtypes are supported by orjson.
    # Fall back to converting to a (possibly nested) Python list.
    return orjson.dumps(content, option=orjson.OPT_SERIALIZE_NUMPY, default=default)



"""Sample on the diffractometer."""

import logging

import numpy as np

from . import util
from .context import TemporaryGeometry
from .engine import Parameter
from .util import Lattice
from .util import libhkl

__all__ = """
    check_lattice
    HklSample
""".split()
logger = logging.getLogger(__name__)


def check_lattice(lattice):
    """Check an Hkl.Lattice for validity

    Raises
    ------
    ValueError
    """

    # TODO: assertion is raised if alpha/beta/gamma are invalid,
    #       which is not propagated back as a Python exception but a segfault.
    a = lattice.a_get()
    b = lattice.b_get()
    c = lattice.c_get()
    alpha = lattice.alpha_get()
    beta = lattice.beta_get()
    gamma = lattice.gamma_get()

    lt = Lattice(a, b, c, alpha, beta, gamma)
    for k, v in lt._asdict().items():
        if v is None:
            raise ValueError(f'Lattice parameter "{k}" unset or invalid')

    lt = Lattice(
        a.value_get(util.units["user"]),
        b.value_get(util.units["user"]),
        c.value_get(util.units["user"]),
        alpha.value_get(util.units["user"]),
        beta.value_get(util.units["user"]),
        gamma.value_get(util.units["user"]),
    )
    logger.debug("Lattice OK: %s", lt)


class HklSample(object):
    """Represents a sample in diffractometer calculations

    Parameters
    ----------
    calc : instance of CalcRecip
        Reciprocal space calculation class
    name : str
        A user-defined name used to refer to the sample
    sample : Hkl.Sample, optional
        A Sample instance from the wrapped Hkl library. Created
        automatically if not specified.
    units : {'user', 'default'}
        Units to use
    lattice : np.ndarray, optional
        The lattice
    U : np.ndarray, optional
        The crystal orientation matrix, U
    UB : np.ndarray, optional
        The UB matrix, where U is the crystal orientation matrix and B is
        the transition matrix of a non-orthonormal (the reciprocal of the
        crystal) in an orthonormal system
    ux : np.ndarray, optional
        ux part of the U matrix
    uy : np.ndarray, optional
        uy part of the U matrix
    uz : np.ndarray, optional
        uz part of the U matrix
    reflections :
        All reflections for the current sample in the form::

            [(h, k, l), ...]

        This assumes the hkl engine is used; generally, the ordered set of
        positions for the engine in-use should be specified.

    PUBLIC API

    .. autosummary::

        ~name
        ~add_reflection
        ~remove_reflection
        ~swap_orientation_reflections
        ~clear_reflections
        ~affine
        ~compute_UB
        ~lattice
        ~reciprocal
        ~reflection_measured_angles
        ~reflection_theoretical_angles
        ~reflections
        ~reflections_details
        ~U
        ~UB
        ~ux
        ~uy
        ~uz

    PRIVATE API

    .. autosummary::

        ~hkl_calc
        ~hkl_sample
        ~__repr__
        ~__str__
        ~_create_reflection
        ~_get_reflection_dict
        ~_refl_matrix
        ~_repr_info
    """

    def __init__(self, calc, sample=None, units="user", **kwargs):
        if sample is None:
            sample = libhkl.Sample.new("")

        self._calc = calc
        self._sample = sample
        self._sample_dict = calc._samples

        self._unit_name = units
        try:
            self._units = util.units[self._unit_name]
        except KeyError:
            raise ValueError(
                f"Unit name '{self._unit_name}' not found.   Allowed names: {list(util.units.keys())}"
            )

        # List of reflections used in computing the U & UB matrices.
        # If UB is computed by refinement of more than two reflections,
        # the code that sets that up will also need to manage this list.
        self._orientation_reflections = []

        for name in "lattice name U UB ux uy uz reflections".split():
            value = kwargs.pop(name, None)
            if value is not None:
                try:
                    setattr(self, name, value)
                except Exception as ex:
                    # These kwargs are funneled down to the gi wrapper
                    # and could raise just about anything. Tack on the
                    # kwarg to help debugging if necessary:
                    ex.message = "%s (attribute=%s)" % (ex, name)
                    raise

        if kwargs:
            raise ValueError("Unsupported kwargs for HklSample: %s" % tuple(kwargs.keys()))

    @property
    def hkl_calc(self):
        """
        The HklCalc instance associated with the sample
        """
        return self._calc

    @property
    def hkl_sample(self):
        """
        The HKL library sample object
        """
        return self._sample

    @property
    def name(self):
        """
        The name of the currently selected sample
        """
        return self._sample.name_get()

    @name.setter
    def name(self, new_name):
        """Replace the current sample

        Parameters
        ----------
        new_name : str
        """
        if new_name in self._sample_dict:
            raise ValueError("Sample with that name already exists")
        sample = self._sample
        old_name = sample.name_get()

        sample.name_set(new_name)

        del self._sample_dict[old_name]
        self._sample_dict[new_name] = self

    @property
    def reciprocal(self):
        """The reciprocal lattice"""
        lattice = self._sample.lattice_get()
        reciprocal = lattice.copy()
        lattice.reciprocal(reciprocal)
        return reciprocal.get(self._units)

    @property
    def lattice(self):
        """The lattice (a, b, c, alpha, beta, gamma)

        a, b, c [nm]
        alpha, beta, gamma [deg]
        """
        lattice = self._sample.lattice_get()
        lattice = lattice.get(self._units)
        return Lattice(*lattice)

    @lattice.setter
    def lattice(self, lattice):
        if not isinstance(lattice, libhkl.Lattice):
            a, b, c, alpha, beta, gamma = lattice
            alpha, beta, gamma = np.radians((alpha, beta, gamma))
            lattice = libhkl.Lattice.new(a, b, c, alpha, beta, gamma)

        check_lattice(lattice)
        self._sample.lattice_set(lattice)
        # TODO: notes mention that lattice should not change, but is it alright
        #       if init() is called again? or should reflections be cleared,
        #       etc?

    @property
    def U(self):
        """
        The crystal orientation matrix, U
        """
        return util.to_numpy(self._sample.U_get())

    @U.setter
    def U(self, new_u):
        self._orientation_reflections = []
        self._sample.U_set(util.to_hkl(new_u))

    def _get_parameter(self, param):
        return Parameter(param, units=self._unit_name)

    @property
    def ux(self):
        """
        ux part of the U matrix
        """
        return self._get_parameter(self._sample.ux_get())

    @property
    def uy(self):
        """
        uy part of the U matrix
        """
        return self._get_parameter(self._sample.uy_get())

    @property
    def uz(self):
        """
        uz part of the U matrix
        """
        return self._get_parameter(self._sample.uz_get())

    @property
    def UB(self):
        """
        The UB matrix, where U is the crystal orientation matrix and B is the
        transition matrix of a non-orthonormal (the reciprocal of the crystal)
        in an orthonormal system

        If written to, the B matrix will be kept constant:
            U * B = UB -> U = UB * B^-1
        """
        return util.to_numpy(self._sample.UB_get())

    @UB.setter
    def UB(self, new_ub):
        self._orientation_reflections = []
        self._sample.UB_set(util.to_hkl(new_ub))

    def _create_reflection(self, h, k, l, detector=None):
        """
        Create a new reflection with the current geometry/detector
        """
        if detector is None:
            detector = self._calc._detector

        return libhkl.SampleReflection.new(self._calc._geometry, detector, h, k, l)

    def compute_UB(self, r1, r2):
        """Compute the UB matrix with two reflections.

        Using the Busing and Levy method, compute the UB matrix for two sample
        reflections, r1 and r2.

        Returns the UB matrix or raises gi.repository.GLib.GError
        (a change from 0.3.15 and before).  Returns ``None`` if no error
        raised but computation was not successful.

        Parameters
        ----------
        r1 : HklReflection
            Reflection 1
        r2 : HklReflection
            Reflection 2

        Returns
        -------
        UB matrix or raises ``gi.repository.GLib.GError``
        """
        if self._sample.compute_UB_busing_levy(r1, r2):
            # TODO: this list defines the order of the orientation reflections
            self._orientation_reflections = [r1, r2]
            return self.UB

    @property
    def reflections(self):
        """
        All reflections for the current sample in the form:
            [(h, k, l), ...]
        """
        return [refl.hkl_get() for refl in self._sample.reflections_get()]

    @reflections.setter
    def reflections(self, refls):
        self.clear_reflections()
        self._orientation_reflections = []
        for refl in refls:
            self.add_reflection(*refl)

    def add_reflection(self, h: float, k: float, l: float, position=None, detector=None, compute_ub: bool = False):
        """Add a reflection, optionally specifying the detector to use

        Parameters
        ----------
        h : (int, float)
            Reflection h
        k : (int, float)
            Reflection k
        l : (int, float)
            Reflection l
        detector : Hkl.Detector, optional
            The detector
        position : list, tuple, or namedtuple, optional
            The physical motor position that this reflection corresponds to
            If not specified, the current geometry of the calculation engine is
            assumed.
        compute_ub : bool, optional
            Calculate the UB matrix with the last two reflections
        """
        calc = self._calc
        if detector is None:
            detector = calc._detector

        if compute_ub and len(self.reflections) < 1:
            raise RuntimeError("Cannot calculate the UB matrix with less than two reflections")

        if compute_ub:
            r1 = self._sample.reflections_get()[-1]

        with TemporaryGeometry(calc):

            def has_valid_position(pos):
                """Raise if invalid, otherwise return boolean."""
                if pos is None:
                    # so use the current motor positions
                    return False
                elif type(pos).__name__.startswith("Pos") or type(pos).__name__.endswith("RealPos"):
                    # This is (probably) a calc.Position namedtuple
                    if False in [isinstance(v, (int, float)) for v in pos]:
                        raise TypeError(f"All values must be numeric, received {pos!r}")
                    if pos._fields != tuple(calc.physical_axis_names):
                        # fmt: off
                        raise KeyError(
                            f"Wrong axes names.  Expected {calc.physical_axis_names},"
                            f" received {pos._fields}"
                        )
                        # fmt: on
                    return True
                elif type(pos).__name__ in "list tuple".split():
                    # note: isinstance(pos, (list, tuple)) includes namedtuple
                    if len(pos) != len(calc.physical_axis_names):
                        # fmt: off
                        raise ValueError(
                            f"Expected {len(calc.physical_axis_names)}"
                            f" positions, received {pos!r}"
                        )
                        # fmt: on
                    if False in [isinstance(v, (int, float)) for v in pos]:
                        raise TypeError(f"All values must be numeric, received {pos!r}")
                    return True
                elif isinstance(pos, (int, float)):
                    raise TypeError(f"Expected positions, received {pos!r}")
                # fmt: off
                raise TypeError(
                    f"Expected list, tuple, or calc.Position() object,"
                    f" received {pos!r}"
                )
                # fmt: on

            if has_valid_position(position):
                calc.physical_positions = position
            r2 = self._sample.add_reflection(calc._geometry, detector, h, k, l)

        if compute_ub:
            self.compute_UB(r1, r2)

        return r2

    def remove_reflection(self, refl):
        """Remove a specific reflection"""
        if not isinstance(refl, libhkl.SampleReflection):
            index = self.reflections.index(refl)
            refl = self._sample.reflections_get()[index]

        return self._sample.del_reflection(refl)

    def clear_reflections(self):
        """Clear all reflections for the current sample."""
        reflections = self._sample.reflections_get()
        for refl in reflections:
            self._sample.del_reflection(refl)
        self._orientation_reflections = []

    def _refl_matrix(self, fcn):
        """Get a reflection angle matrix."""
        sample = self._sample
        refl = sample.reflections_get()
        refl_matrix = np.zeros((len(refl), len(refl)))

        for i, r1 in enumerate(refl):
            for j, r2 in enumerate(refl):
                if i != j:
                    refl_matrix[i, j] = fcn(r1, r2)

        return refl_matrix

    @property
    def reflection_measured_angles(self):
        return self._refl_matrix(self._sample.get_reflection_measured_angle)

    @property
    def reflection_theoretical_angles(self):
        return self._refl_matrix(self._sample.get_reflection_theoretical_angle)

    def affine(self):
        """
        Refine (affine) the sample lattice parameters from the list of reflections.
        """
        return self._sample.affine()

    def _repr_info(self):
        r = [
            f"name={self.name!r}",
            f"lattice={self.lattice!r}",
            f"ux={self.ux!r}",
            f"uy={self.uy!r}",
            f"uz={self.uz!r}",
            f"U={self.U!r}",
            f"UB={self.UB!r}",
            f"reflections={self.reflections!r}",
        ]

        return r

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._repr_info())})"

    def __str__(self):
        info = self._repr_info()
        info.append(f"reflection_measured_angles={self.reflection_measured_angles!r}")
        info.append(f"reflection_theoretical_angles={self.reflection_theoretical_angles!r}")
        return f"{self.__class__.__name__}({', '.join(info)}))"

    def _get_reflection_dict(self, refl):
        """Return dictionary with reflection details."""
        h, k, l = refl.hkl_get()
        geom = refl.geometry_get()
        return {
            "reflection": {"h": h, "k": k, "l": l},
            "flag": refl.flag_get(),  # not used by hklpy
            "wavelength": geom.wavelength_get(1),
            "position": dict(zip(geom.axis_names_get(), geom.axis_values_get(1))),
            "orientation_reflection": refl in self._orientation_reflections,
        }

    @property
    def reflections_details(self):
        """
        Return a list with details of all reflections.

        NOTE: reflections_details() uses the canonical
        names for the real positioners.  The mapping
        to physical axis names happens in :mod`hkl.calc`.
        """
        refls = self._sample.reflections_get()
        for r in self._orientation_reflections:
            if r not in refls:
                # Edge case when orientation reflection was
                # deleted from the list in libhkl.
                refls.append(r)
        return [self._get_reflection_dict(r) for r in refls]

    def swap_orientation_reflections(self):
        """Swap the 2 [UB] reflections, re-compute & return new [UB]."""
        if len(self._orientation_reflections) != 2:
            raise ValueError(
                "Must have exactly 2 orientation reflections defined"
                " in order to make a swap and re-compute [UB]."
                f"  There are {len(self._orientation_reflections)}"
                " orientation reflection(s) defined now."
            )
        refls = self._orientation_reflections
        return self.compute_UB(*refls[::-1])



"""
Save and restore Diffractometer Configuration.

PUBLIC API

.. autosummary::

    ~DiffractometerConfiguration

PRIVATE API

note: DC classes: Diffractometer Configuration

.. autosummary::

    ~_check_key
    ~_check_not_value
    ~_check_range
    ~_check_type
    ~_check_value
    ~DCConstraint
    ~DCLattice
    ~DCReflection
    ~DCSample
    ~DCConfiguration

.. note:: New in v1.1.
"""

__all__ = [
    "DiffractometerConfiguration",
]

import datetime
import json
import pathlib
import typing
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import List

import numpy
import pyRestTable
import yaml
from apischema import deserialize
from apischema import serialize

from .util import libhkl

# TODO: can these be learned from the diffractometer? (not constraints but axes?)
AX_MIN = -360.0  # lowest allowed value for real-space axis
AX_MAX = 360.0  # highest allowed value for real-space axis

DEFAULT_WAVELENGTH = 1.54  # angstrom
EXPORT_FORMATS = "dict json yaml".split()

SIGNIFICANT_DIGITS = 7


# standard value checks, raise exception(s) when appropriate
def _check_key(key, biblio, intro):
    """(internal) Raise KeyError if key is not in biblio."""
    if key not in biblio:
        raise KeyError(f"{intro}:  expected {key!r} not in {biblio}")


def _check_not_value(actual, avoid, intro):
    """(internal) Raise ValueError if actual IS equal to expected."""
    if actual == avoid:
        raise ValueError(f"{intro}:  received: {actual}  cannot be: {avoid}")


def _check_range(value, low, high, intro):
    """(internal) Raise ValueError if value is not between low & high."""
    if low > high:
        raise ValueError(f"{intro}:  {low} should not be greater than {high}")
    if not low <= value <= high:
        raise ValueError(f"{intro}:  {value} is not between {low} & {high}")


def _check_type(actual, expected, intro):
    """(internal) Raise TypeError if actual is not an instance of expected."""
    if not isinstance(actual, expected):
        raise TypeError(f"{intro}:  received: {actual}  expected: {expected}")


def _check_value(actual, expected, intro):
    """(internal) Raise ValueError if actual is not equal to expected."""
    if actual != expected:
        raise ValueError(f"{intro}:  received: {actual}  expected: {expected}")


@dataclass
class DCConstraint:
    """
    (internal) Configuration of one diffractometer axis constraint.
    """

    low_limit: float
    """
    Lowest acceptable value for this axis when computing real-space solutions
    from given reciprocal-space positions.
    """

    high_limit: float
    """
    Highest acceptable value for this axis when computing real-space solutions
    from given reciprocal-space positions.
    """

    value: float
    """
    Constant value used (on condition) for ``forward(hkl)`` calculation.

    Implemented by diffractometer :attr:`~hkl.engine.Engine.mode`.

    The diffractometer engine's :attr:`~hkl.engine.Engine.mode` (such as E4CV's
    ``constant_phi`` mode) controls whether or not the axis is to be held
    constant.
    """

    fit: bool = True
    """
    (deprecated) Not used as a constraint.  Value is ignored. See
    :class:`~hkl.util.Constraint`.
    """

    @property
    def values(self):
        """Return the list of values in order."""
        return list(self.__dict__.values())

    def validate(self, cname):
        """
        Check this constraint has values the diffractometer can accept.

        Assumes diffractometer real axis limits of AX_MIN to AX_MAX degrees.
        """
        _check_range(self.low_limit, AX_MIN, AX_MAX, f"{cname} low_limit")
        _check_range(self.high_limit, AX_MIN, AX_MAX, f"{cname} high_limit")
        _check_range(self.value, AX_MIN, AX_MAX, f"{cname} value")
        # no additional validation needed for 'fit'


@dataclass
class DCLattice:
    """(internal) Configuration of one crystal lattice."""

    a: float
    """unit cell length :math:`a` (angstrom)"""

    b: float
    """unit cell length :math:`b` (angstrom)"""

    c: float
    """unit cell length :math:`c` (angstrom)"""

    alpha: float
    """unit cell angle alpha (degrees)"""

    beta: float
    """unit cell angle beta (degrees)"""

    gamma: float
    """unit cell angle gamma (degrees)"""

    def validate(self, *_args):
        """Check this lattice has values the diffractometer can accept."""
        _check_range(self.a, 1e-6, 1e6, "a")
        _check_range(self.b, 1e-6, 1e6, "b")
        _check_range(self.c, 1e-6, 1e6, "c")
        _check_range(self.alpha, 1e-6, 180.0 - 1e-6, "alpha")
        _check_range(self.beta, 1e-6, 180.0 - 1e-6, "beta")
        _check_range(self.gamma, 1e-6, 180.0 - 1e-6, "gamma")
        # exclude zero
        _check_not_value(self.alpha, 0.0, "alpha")
        _check_not_value(self.beta, 0.0, "beta")
        _check_not_value(self.gamma, 0.0, "gamma")

    @property
    def values(self):
        """Return the list of values in order."""
        return list(self.__dict__.values())


@dataclass
class DCReflection:
    """(internal) Configuration of one orientation reflection."""

    reflection: Dict[str, float]
    """
    Reciprocal-space axis positions.  Keys must match in the list of
    ``reciprocal_axes``.
    """

    position: Dict[str, float]
    """
    Real-space axis positions.  Keys must match in the list of
    ``canonical_axes``.
    """

    wavelength: float
    """Wavelength (angstroms) at which *this* reflection was measured."""

    orientation_reflection: bool
    """Use this reflection for calculating :math:`UB` matrix?"""

    flag: int = 1
    """(only used by *libhkl*)"""

    def validate(self, dc_obj):
        """Check this reflection has values the diffractometer can accept."""
        _check_range(self.wavelength, 1e-6, 1e6, "wavelength")
        # physics: reciprocal-space won't stretch any further
        q_max = 4 * numpy.pi / self.wavelength
        q_min = -q_max
        for axis, value in self.reflection.items():
            _check_key(axis, dc_obj.reciprocal_axes_names, f"reciprocal-space axis {axis}")
            _check_range(value, q_min, q_max, f"reciprocal-space axis {axis}")
        for axis, value in self.position.items():
            _check_key(axis, dc_obj.canonical_axes_names, f"real-space axis {axis}")
            _check_range(value, AX_MIN, AX_MAX, f"real-space axis {axis}")
        # do not validate 'flag' (not used in hklpy)


@dataclass
class DCSample:
    """(internal) Configuration of one crystalline sample with a lattice."""

    name: str
    """Name of this crystalline sample."""

    lattice: DCLattice
    """Crystal lattice parameters (angstroms and degrees)"""

    reflections: List[DCReflection]
    """List of orientation reflections."""

    UB: List[List[float]]
    """
    Orientation matrix (3 x 3).  U is the crystal orientation matrix relative
    to the diffractometer and B is the transition matrix of a non-orthonormal
    (the reciprocal of the crystal) in an orthonormal system.
    """

    # TODO: Once py38 is dropped, re-enable the default value setting
    U: List[List[float]]  # = field(default_factory=list[list[float]])
    """
    Orientation matrix (3 x 3) of the crystal relative to the diffractometer.
    (optional)
    """

    def validate(self, dc_obj):
        """Check this sample has values the diffractometer can accept."""
        self.lattice.validate()
        _check_not_value(self.name.strip(), "", "name cannot be empty")
        for reflection in self.reflections:
            reflection.validate(dc_obj)
        for k in "U UB".split():
            arr = numpy.array(getattr(self, k))
            _check_value(arr.shape, (3, 3), f"{k} matrix shape")
            for i in range(len(arr)):  # Want i, j for reporting
                for j in range(len(arr[i])):
                    _check_type(arr[i][j], (float, numpy.floating), f"{k}[{i}][{j}]")

    def write(self, diffractometer):
        """Write sample details to diffractometer."""
        sample = diffractometer.calc._samples.get(self.name)
        lattice_parameters = list(self.lattice.values)
        if sample is None:
            sample = diffractometer.calc.new_sample(self.name, lattice=lattice_parameters)
        else:
            sample.lattice = lattice_parameters

        reflection_list = []
        for reflection in self.reflections:
            rdict = asdict(reflection)
            # fmt: off
            args = [  # hkl values
                *list(rdict["reflection"].values()),
                tuple(rdict["position"].values())]
            # fmt: on

            # temporarily, change the wavelength
            w0 = diffractometer.calc.wavelength
            w1 = rdict["wavelength"]
            try:
                diffractometer.calc.wavelength = w1
                r = sample.add_reflection(*args)
                if rdict["orientation_reflection"]:
                    reflection_list.append(r)
            except RuntimeError as exc:
                raise RuntimeError(f"could not add reflection({args}, wavelength={w1})") from exc
            finally:
                diffractometer.calc.wavelength = w0
            # Remaining code will not be executed if exception was raised.

            if len(reflection_list) > 1:
                r1, r2 = reflection_list[0], reflection_list[1]
                sample.compute_UB(r1, r2)


@dataclass
class DCConfiguration:
    """
    (internal) Full structure of the diffractometer configuration.

    Optional (keyword) attributes are not used to restore a diffractometer's
    configuration.

    Required (non-optional) attributes are used by ``restore()`` to either match
    the diffractometer or restore the configuration.
    """

    geometry: str
    """
    Name of the diffractometer geometry as provided by the back-end
    computation library.  MUST match diffractometer to restore.
    """

    engine: str
    """
    Name of the computational support for the reciprocal-space (pseudo) axes.
    MUST match in the list provided by the diffractometer geometry to restore.
    The *engine* defines the list of the reciprocal-space (pseudo) axes.
    """

    library: str
    """
    Name of the back-end computation library.  MUST match diffractometer to
    restore.
    """

    mode: str
    """
    Diffractometer calculation mode.  Chosen from list provided by the
    back-end computation library.  MUST match in the list provided by the
    diffractometer to restore.
    """

    canonical_axes: List[str]
    """
    List of the diffractometer real-space axis names.  Both the exact spelling
    and order are defined by the back-end computation library.  MUST match
    diffractometer to restore.
    """

    real_axes: List[str]
    """
    User-defined real-space axis names. MUST match diffractometer to restore.
    The length and order of this list must be the same as the
    ``canonical_axes``. It is used to resolve any (real-space) ``positioner``
    names in this file.
    """

    reciprocal_axes: List[str]
    """
    List of names of the diffractometer reciprocal-space (pseudo) axes. Both
    the exact spelling and order are defined by the back-end computation
    library ``engine``.
    MUST match diffractometer to restore.
    """

    constraints: Dict[str, DCConstraint]
    """
    Limits to be imposed on the real-space axes for operations and
    computations.  Keys must match in the list of ``canonical_axes``.
    """

    samples: Dict[str, DCSample]
    """
    Crystalline samples (lattice and orientation reflections).
    The sample name is used as the key in the dictionary.
    """

    # -------------------- optional attributes
    name: str = ""
    """
    Name of this diffractometer. (optional)
    """

    datetime: str = ""
    """
    Date and time this configuration was recorded. (optional)
    """

    wavelength_angstrom: float = field(default_factory=float)
    """
    Wavelength (angstrom) of the incident radiation. (optional)
    """

    energy_keV: float = field(default_factory=float)
    """
    Energy (keV) of the incident beam.  Useful for synchrotron X-ray
    instruments. (optional)
    """

    hklpy_version: str = ""
    """
    Version of the *hklpy* Python package used to create this diffractometer
    configuration content. (optional)
    """

    library_version: str = ""
    """
    Version information of the back-end computation library. (optional)
    """

    python_class: str = ""
    """
    Name of the Python class that defines this diffractometer. (optional)
    """

    other: Dict[str, typing.Any] = field(default_factory=dict)
    """
    *Any* other content goes into this dictionary (comments, unanticipated
    keys, ...) (optional)
    """

    def validate(self, dc_obj):
        """
        Check this configuration has values the diffractometer can accept.

        PARAMETERS

        dc_obj *DiffractometerConfiguration*:
            The DiffractometerConfiguration object.
        """
        diffractometer = dc_obj.diffractometer
        _check_value(self.geometry, diffractometer.calc._geometry.name_get(), "geometry")
        _check_key(self.engine, diffractometer.calc._engine_names, "engine")
        _check_value(self.engine, diffractometer.calc.engine.name, "engine")
        _check_value(self.library, libhkl.__name__, "library")
        _check_value(self.canonical_axes, dc_obj.canonical_axes_names, "canonical_axes")
        _check_value(self.reciprocal_axes, dc_obj.reciprocal_axes_names, "reciprocal_axes")
        _check_key(self.mode, diffractometer.engine.modes, "mode")

        for k in "canonical real reciprocal".split():
            # number of axes must match
            _check_value(
                len(getattr(self, f"{k}_axes")), len(getattr(dc_obj, f"{k}_axes_names")), f"number of {k}_axes"
            )

        for cname, constraint in self.constraints.items():
            try:
                _check_key(cname, dc_obj.canonical_axes_names, "constraint axis")
            except KeyError:
                _check_key(cname, dc_obj.real_axes_names, "constraint axis")
            constraint.validate(cname)

        for sample in self.samples.values():
            sample.validate(dc_obj)

    def write(self, diffractometer, restore_constraints=True):
        """Update diffractometer with configuration."""
        from .util import Constraint

        # fmt: off
        if not isinstance(restore_constraints, bool):
            raise TypeError(
                "'restore_constraints' must be True or False,"
                f" received {restore_constraints}"
            )
        # fmt: on

        # don't reset the wavelength
        # don't reset the (real-space) positions
        # don't reset the (reciprocal-space) positions
        diffractometer.engine.mode = self.mode

        # fmt: off
        if restore_constraints:
            diffractometer.apply_constraints(
                {
                    k: Constraint(*constraint.values)
                    for k, constraint in self.constraints.items()
                }
            )
        # fmt: on

        for sample in self.samples.values():
            sample.write(diffractometer)


class DiffractometerConfiguration:
    """
    Save and restore Diffractometer Configuration.

    .. autosummary::

        ~export
        ~preview
        ~restore
        ~model
        ~reset_diffractometer
        ~reset_diffractometer_constraints
        ~reset_diffractometer_samples
        ~from_dict
        ~from_json
        ~from_yaml
        ~to_dict
        ~to_json
        ~to_yaml
        ~canonical_axes_names
        ~real_axes_names
        ~reciprocal_axes_names
    """

    def __init__(self, diffractometer):
        from .diffract import Diffractometer

        if not isinstance(diffractometer, Diffractometer):
            raise TypeError("diffractometer should be 'Diffractometer' or subclass.")
        self.diffractometer = diffractometer

    def export(self, fmt="json"):
        """
        Export configuration in a recognized format (dict, JSON, YAML, file).

        PARAMETERS

        fmt *str* or *pathlib.Path* object:
            One of these: ``None``, ``"dict"``, ``"json"``, ``"yaml"``. If
            ``None`` (or empty string or no argument at all), then JSON will be
            the default.
        """
        path = None
        if isinstance(fmt, pathlib.Path):
            path = fmt
            fmt = "json"  # use default format

        fmt = (fmt or "json").lower()
        if fmt == "yml":
            fmt = "yaml"  # a common substitution, just being friendly
        if fmt not in EXPORT_FORMATS:
            raise ValueError(f"fmt must be one of {EXPORT_FORMATS}, received {fmt!r}")
        data = getattr(self, f"to_{fmt}")()
        if path is not None:
            with open(path, "w") as f:
                f.write(data)
        return data

    def preview(self, data, show_constraints=False, show_reflections=False):
        """
        List the samples in the configuration.

        PARAMETERS

        data *dict* or *str* *pathlib.Path* object:
            Structure (dict, json, or yaml) with diffractometer configuration
            or pathlib object referring to a file with one of these formats.
        show_constraints *bool*:
            If ``True`` (default: ``False``), will also show any constraints
            in a separate table.
        show_reflections *bool*:
            If ``True`` (default: ``False``), will also show reflections, if
            any, in a separate table for each sample.
        """
        if isinstance(data, pathlib.Path):
            if not data.exists():
                raise FileNotFoundError(f"{data}")
            with open(data) as f:
                data = f.read()

        if isinstance(data, str):
            if data.strip().startswith("{"):
                data = json.loads(data)
            else:
                data = yaml.load(data, Loader=yaml.Loader)

        return self._preview(data, show_constraints, show_reflections)

    def _preview(self, data, show_constraints=False, show_reflections=False):
        if not isinstance(data, dict):
            raise TypeError(f"Cannot interpret configuration data: {type(data)}")

        def float_format(v):
            return f"{round(v, SIGNIFICANT_DIGITS):g}"

        text = (
            f"name: {data.get('name', '-n/a-')}"
            f"\ndate: {data.get('datetime', '-n/a-')}"
            f"\ngeometry: {data['geometry']}"
        )

        title = "Table of Samples"
        table = pyRestTable.Table()
        table.labels = "# sample a b c alpha beta gamma #refl".split()
        for i, sname in enumerate(data["samples"], start=1):
            sample = data["samples"][sname]
            row = [i, sname]
            for v in sample["lattice"].values():
                row.append(float_format(v))
            row.append(len(sample["reflections"]))
            table.addRow(row)
        text += f"\n\n{title}\n{table}"

        if show_reflections:
            for sname, sample in data["samples"].items():
                if len(sample["reflections"]) == 0:
                    continue  # nothing to report
                title = f"Table of Reflections for Sample: {sname}"
                table = pyRestTable.Table()
                refl = sample["reflections"][0]
                table.addLabel("#")
                table.labels += list(refl["reflection"])
                table.labels += list(refl["position"])
                table.addLabel("wavelength")
                table.addLabel("orient?")
                for i, refl in enumerate(sample["reflections"], start=1):
                    row = [i]
                    row += [float_format(v) for v in refl["reflection"].values()]
                    row += [float_format(v) for v in refl["position"].values()]
                    row.append(f"{round(refl['wavelength'], SIGNIFICANT_DIGITS):g}")
                    row.append(str(refl["orientation_reflection"]))
                    table.addRow(row)
                text += f"\n\n{title}\n{table}"

        if show_constraints and len(data["constraints"]) > 0:
            title = "Table of Axis Constraints"
            table = pyRestTable.Table()
            table.labels = "axis low_limit high_limit value fit?".split()
            for aname, constraint in data["constraints"].items():
                row = [aname]
                for k in "low_limit high_limit value".split():
                    row.append(float_format(constraint[k]))
                row.append(f"{constraint['fit']}")
                table.addRow(row)
            text += f"\n\n{title}\n{table}"

        return text

    def restore(self, data, clear=True, restore_constraints=True):
        """
        Restore configuration from a recognized format (dict, JSON, YAML, file).

        Instead of guessing, recognize the kind of config data by its structure.

        PARAMETERS

        data *dict* or *str* *pathlib.Path* object:
            Structure (dict, json, or yaml) with diffractometer configuration
            or pathlib object referring to a file with one of these formats.
        clear *bool*:
            If ``True`` (default), remove any previous configuration of the
            diffractometer and reset it to default values before restoring the
            configuration.

            If ``False``, sample reflections will be append with all reflections
            included in the configuration data for that sample.  Existing
            reflections will not be changed.  The user may need to edit the
            list of reflections after ``restore(clear=False)``.
        restore_constraints *bool*:
            If ``True`` (default), restore any constraints provided.

        Note: Can't name this method "import", it's a reserved Python word.
        """
        importer = None

        if not isinstance(clear, bool):
            raise TypeError(f"clear must be either True or False, received {clear}")

        if isinstance(data, pathlib.Path):
            if not data.exists():
                raise FileNotFoundError(f"{data}")
            with open(data) as f:
                data = f.read()

        if isinstance(data, dict):
            importer = self.from_dict
        elif isinstance(data, str):
            if data.strip().startswith("{"):
                importer = self.from_json
            else:
                importer = self.from_yaml
        if importer is None:
            raise TypeError("Unrecognized configuration structure.")

        importer(data, clear=clear, restore_constraints=restore_constraints)

    @property
    def canonical_axes_names(self):
        """Names of the real-space axes, defined by the back-end library."""
        return self.diffractometer.calc._geometry.axis_names_get()

    @property
    def real_axes_names(self):
        """Names of the real-space axes, defined by the user."""
        return list(self.diffractometer.RealPosition._fields)

    @property
    def reciprocal_axes_names(self):
        """Names of the reciprocal-space axes, defined by the back-end library."""
        return list(self.diffractometer.PseudoPosition._fields)

    @property
    def model(self) -> DCConfiguration:
        """Return validated diffractometer configuration object."""
        diffractometer = self.diffractometer

        # either an empty dict or maps renamed axes to canonical
        xref = dict(diffractometer.calc._axis_name_to_original)

        def canonical_name(axis):
            return xref.get(axis, axis)

        data = {
            "name": diffractometer.name,
            "geometry": diffractometer.calc._geometry.name_get(),
            "datetime": str(datetime.datetime.now()),
            "python_class": diffractometer.__class__.__name__,
            "engine": diffractometer.calc.engine.name,
            "mode": diffractometer.calc.engine.mode,
            "canonical_axes": self.canonical_axes_names,
            "real_axes": self.real_axes_names,
            "reciprocal_axes": self.reciprocal_axes_names,
            "energy_keV": diffractometer.calc.energy,  # for X-ray instruments
            "wavelength_angstrom": diffractometer.calc.wavelength,
            "hklpy_version": diffractometer._hklpy_version_,
            "library": libhkl.__name__,
            "library_version": libhkl.VERSION,
            # fmt: off
            "constraints": {
                canonical_name(axis): {
                    parm: getattr(constraint, parm)
                    for parm in "low_limit high_limit value fit".split()
                }
                for axis, constraint in diffractometer._constraints_dict.items()
            },
            # fmt: on
            "samples": {
                sname: {
                    "name": sample.name,
                    "lattice": sample.lattice._asdict(),
                    "reflections": sample.reflections_details,
                    "U": sample.U.tolist(),
                    "UB": sample.UB.tolist(),
                }
                for sname, sample in diffractometer.calc._samples.items()
            },
        }
        obj = deserialize(DCConfiguration, data)  # also validates structure
        obj.validate(self)  # check that values are valid
        return obj

    def reset_diffractometer(self):
        """Reset the diffractometer to the default configuration."""
        self.diffractometer.wavelength = DEFAULT_WAVELENGTH
        self.diffractometer.engine.mode = self.diffractometer.engine.modes[0]
        self.reset_diffractometer_constraints()
        self.reset_diffractometer_samples()

    def reset_diffractometer_constraints(self):
        """Reset the diffractometer constraints to defaults."""
        from .util import Constraint

        # fmt: off
        self.diffractometer._set_constraints(
            {
                k: Constraint(-180., 180., 0.0, True)
                for k in self.real_axes_names
            }
        )
        # fmt: on

    def reset_diffractometer_samples(self):
        """Reset the diffractometer sample dict to defaults."""
        for k in list(self.diffractometer.calc._samples):
            self.diffractometer.calc._samples.pop(k)

        # fmt: off
        a0 = DEFAULT_WAVELENGTH  # coincidentally
        self.diffractometer.calc.new_sample(
            "main", lattice=(a0, a0, a0, 90., 90., 90.)
        )
        # fmt: on

    def from_dict(self, data, clear=True, restore_constraints=True):
        """
        Load diffractometer configuration from Python dictionary.

        PARAMETERS

        data *dict*:
            structure (dict) with diffractometer configuration
        clear *bool*:
            If ``True`` (default), remove any previous configuration of the
            diffractometer and reset it to default values before restoring the
            configuration.
        """
        # note: deserialize first runs a structural validation
        model = deserialize(DCConfiguration, data)
        model.validate(self)  # check that values are valid
        if clear:
            self.reset_diffractometer()
        # tell the model to update the diffractometer
        model.write(self.diffractometer, restore_constraints=restore_constraints)

    def to_dict(self):
        """Report diffractometer configuration as Python dictionary."""
        return serialize(DCConfiguration, self.model)

    def from_json(self, data, clear=True, restore_constraints=True):
        """
        Load diffractometer configuration from JSON text.

        PARAMETERS

        data *str* (JSON):
            structure (JSON string) with diffractometer configuration
        clear *bool*:
            If ``True`` (default), remove any previous configuration of the
            diffractometer and reset it to default values before restoring the
            configuration.
        """
        self.from_dict(json.loads(data), clear=clear, restore_constraints=restore_constraints)

    def to_json(self, indent=2):
        """Report diffractometer configuration as JSON text."""
        return json.dumps(self.to_dict(), indent=indent)

    def from_yaml(self, data, clear=True, restore_constraints=True):
        """
        Load diffractometer configuration from YAML text.

        PARAMETERS

        data *str* (YAML):
            structure (YAML string) with diffractometer configuration
        clear *bool*:
            If ``True`` (default), remove any previous configuration of the
            diffractometer and reset it to default values before restoring the
            configuration.
        """
        # fmt: off
        self.from_dict(
            yaml.load(data, Loader=yaml.Loader),
            clear=clear,
            restore_constraints=restore_constraints
        )
        # fmt: on

    def to_yaml(self, indent=2):
        """
        Report diffractometer configuration as YAML text.

        Order of appearance may be important for some entries, such as the list
        of reflections.  Use ``sort_keys=False`` here. Don't make ``sort_keys``
        a keyword argument that could be changed.
        """
        return yaml.dump(self.to_dict(), indent=indent, sort_keys=False)



class UsingEngine(object):
    """Context manager that uses a calculation engine temporarily"""

    def __init__(self, calc, engine):
        self.calc = calc
        self.engine = engine
        self.old_engine = None

    def __enter__(self):
        self.old_engine = self.calc.engine
        if self.engine is not None:
            self.calc.engine = self.engine

    def __exit__(self, type_, value, traceback):
        if self.engine is not None:
            self.calc.engine = self.old_engine


class TemporaryGeometry(object):
    """Context manager that restores physical geometry after a block of code"""

    def __init__(self, calc):
        self.calc = calc

    def __enter__(self):
        self.geometry = self.calc._geometry.copy()

    def __exit__(self, type_, value, traceback):
        self.calc._geometry = self.geometry
        self.calc._re_init()



import pytest
from ophyd import Component
from ophyd import Device
from ophyd import EpicsMotor
from ophyd import EpicsSignal
from ophyd import Kind

import hkl

from .common import IOC_PV_PREFIX_GP

OMEGA_PV = f"{IOC_PV_PREFIX_GP}m1"
CHI_PV = f"{IOC_PV_PREFIX_GP}m2"
PHI_PV = f"{IOC_PV_PREFIX_GP}m3"
TTH_PV = f"{IOC_PV_PREFIX_GP}m4"

MOTOR_RECORD_CONFIG_ATTRS = sorted(
    """
        acceleration
        motor_egu
        user_offset
        user_offset_dir
        velocity
    """.split()
)


class FourC(hkl.SimMixin, hkl.E4CV):
    omega = Component(EpicsMotor, OMEGA_PV)
    chi = Component(EpicsMotor, CHI_PV)
    phi = Component(EpicsMotor, PHI_PV)
    tth = Component(EpicsMotor, TTH_PV)


class MyDevice(Device):
    omega = Component(EpicsMotor, OMEGA_PV)


device = MyDevice("", name="device")
fourc = FourC("", name="fourc")
motor = EpicsMotor(OMEGA_PV, name="motor")


@pytest.mark.parametrize("attr", MOTOR_RECORD_CONFIG_ATTRS)
@pytest.mark.parametrize(
    "motor, parent",
    [
        [motor, motor],
        [device.omega, device],
        [fourc.omega, fourc],
        [fourc.chi, fourc],
        [fourc.phi, fourc],
        [fourc.tth, fourc],
    ],
)
def test_i325(attr, motor, parent):
    """
    Check the configuration_attrs for EpicsMotor itself and as Component.

    https://github.com/bluesky/hklpy/issues/325

    EPICS IOC and motor records are necessary to demonstrate this problem. We
    prove the problem is NOT in the ophyd Devices by including them in the
    tests.

    Comparison of device's ``.read_configuration()`` output can show this
    problem. Compare EpicsMotor with any/all Diffractometer motors.
    """
    parent.wait_for_connection()
    motor.wait_for_connection()

    device_configuration = parent.read_configuration()
    device_configuration_keys = sorted(list(device_configuration))
    expected_key = f"{motor.name}_{attr}"
    assert attr in motor.component_names, f"{expected_key=!r}"
    component = getattr(motor, attr)
    assert isinstance(component, EpicsSignal)
    assert component.kind == Kind.config, f"{component.kind=!r}"

    assert expected_key in device_configuration_keys, f"{expected_key=!r}"
    assert sorted(motor.configuration_attrs) == MOTOR_RECORD_CONFIG_ATTRS, f"{motor.name=!r}"



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# flake8: noqa

import pathlib
import sys

docs_source = pathlib.Path(__file__).parent
sys.path.insert(0, str(docs_source.parent.parent))

# imports here for sphinx to build the documents without many WARNINGS.
import hkl
import hkl.calc
import hkl.diffract
import hkl.geometries
import hkl.user

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "hklpy"
copyright = "2014-2023, Brookhaven National Laboratory"
# author = 'Bluesky team'
release = hkl.__version__
version = ".".join(release.split(".")[:2])
short_version = version
if "+g" in version and ".d2" in version:
    # Extra date (1.0.5.dev146+gbb34ee0.d20231102) makes the title too long.
    short_version = short_version.rsplit(".d2", 1)[0]
today_fmt = "%Y-%m-%d %H:%M"
html_title = f"{project} {short_version}"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "nbsphinx",
]
extensions.append("sphinx_tabs.tabs")  # this must be last

templates_path = ["_templates"]
exclude_patterns = ["**.ipynb_checkpoints"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_title = f"{project} {version}"
html_static_path = ["_static"]


# -- Options for autodoc ---------------------------------------------------

autodoc_exclude_members = ",".join(
    """
    __weakref__
    _component_kinds
    _device_tuple
    _required_for_connection
    _sig_attrs
    _sub_devices
    calc_class
    component_names
    """.split()
)
autodoc_default_options = {
    # 'members': 'var1, var2',
    # 'member-order': 'bysource',
    "private-members": True,
    # "special-members": "__init__",
    # 'undoc-members': True,
    "exclude-members": autodoc_exclude_members,
}
autodoc_mock_imports = [
    "pint",
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}
inheritance_graph_attrs = {"rankdir": "LR"}
inheritance_node_attrs = {"fontsize": 24}
autosummary_generate = True



"""
Low-level support to Hkl computation engine support for diffractometers

.. autosummary::

    ~CalcParameter
    ~Engine
    ~Parameter
    ~Solution

"""

from __future__ import print_function

import logging
from collections import OrderedDict

from . import util

__all__ = """
    CalcParameter
    Engine
    Parameter
    Solution
""".split()
logger = logging.getLogger(__name__)


AXES_READ = 0
AXES_WRITTEN = 1


class Parameter(object):
    """HKL library parameter object

    .. autosummary::

        ~inverted
        ~hkl_parameter
        ~units
        ~name
        ~value
        ~user_units
        ~default_units
        ~limits

    Example::

        Parameter(
                name='None (internally: ux)',
                limits=(min=-180.0, max=180.0),
                value=0.0,
                fit=True,
                inverted=False,
                units='Degree')

    """

    def __init__(self, param, units="user", name=None, inverted=False):
        self._param = param
        self._unit_name = units
        self._units = util.units[units]
        self._name = name
        self._inverted = inverted

    @property
    def inverted(self):
        """Is the value inverted internally?"""
        return self._inverted

    @property
    def hkl_parameter(self):
        """The HKL library parameter object"""
        return self._param

    @property
    def units(self):
        """Either ``user`` or ``default``."""
        return self._unit_name

    @property
    def name(self):
        """Name of this parameter."""
        name = self._param.name_get()
        if self._name != name:
            return f"{self._name} (internally: {name})"
        return name

    @property
    def value(self):
        """
        Value used (in :meth:`~hkl.calc.CalcRecip.forward()` method) if held
        constant by the :attr:`~hkl.engine.Engine.mode`.
        """
        return self._param.value_get(self._units)

    @value.setter
    def value(self, value):
        if self._inverted:
            value *= -1.0

        self._param.value_set(value, self._units)

    @property
    def user_units(self):
        """A string representing the user unit type"""
        return self._param.user_unit_get()

    @property
    def default_units(self):
        """A string representing the default unit type"""
        return self._param.default_unit_get()

    @property
    def fit(self):
        """True if the parameter can be fit or not"""
        return bool(self._param.fit_get())

    @fit.setter
    def fit(self, fit):
        self._param.fit_set(int(fit))

    @property
    def limits(self):
        """(low, high)"""
        if self._inverted:
            low, high = self._param.min_max_get(self._units)
            return (-high, -low)
        else:
            return self._param.min_max_get(self._units)

    @limits.setter
    def limits(self, lims):
        low, high = lims
        if self._inverted:
            self._param.min_max_set(-high, -low, self._units)
        else:
            self._param.min_max_set(low, high, self._units)

    def _repr_info(self):
        r = [
            f"name={self.name!r}",
            f"limits={self.limits!r}",
            f"value={self.value!r}",
            f"fit={self.fit!r}",
            f"inverted={self.inverted!r}",
        ]

        if self._unit_name == "user":
            r.append(f"units={self.user_units!r}")
        else:
            r.append(f"units={self.default_units!r}")

        return r

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._repr_info())})"

    def __str__(self):
        info = self._repr_info()
        # info.append(self.)
        return f"{self.__class__.__name__}({', '.join(info)})"


class Solution(object):
    """
    Solution of the conversion from (hkl) to axes.

    .. autosummary::

        ~axis_names
        ~positions
        ~select
        ~units
    """

    def __init__(self, engine, list_item, class_):
        self._class = class_
        self._list_item = list_item.copy()
        self._geometry = list_item.geometry_get().copy()
        self._engine = engine

    def __getitem__(self, axis):
        return self._geometry.axis_get(axis)

    @property
    def axis_names(self):
        """List real axis names of the solution."""
        return self._geometry.axis_names_get()

    @property
    def positions(self):
        """List real axis values of the solution."""
        return self._class(*self._geometry.axis_values_get(self._engine._units))

    @property
    def units(self):
        """ """
        return self._engine.units

    def select(self):
        """Select a solution."""
        self._engine._engine_list.select_solution(self._list_item)

    def _repr_info(self):
        r = [
            repr(self.positions),
            f"units={self._engine.units!r}",
        ]

        return r

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._repr_info())})"


class Engine(object):
    """
    HKL calculation engine

    .. autosummary::

        ~engine
        ~mode
        ~name
        ~parameters
        ~parameters_values
        ~pseudo_axes
        ~pseudo_axis_names
        ~pseudo_positions
        ~solutions
        ~units
        ~update
    """

    def __init__(self, calc, engine, engine_list):
        self._calc = calc
        self._engine = engine
        self._engine_list = engine_list

    @property
    def name(self):
        """Name of this engine."""
        return self._engine.name_get()

    @property
    def axes_c(self):
        """HKL real axis names (held constant during forward() computation)."""
        return [axis for axis in self.axes_r if axis not in self.axes_w]

    @property
    def axes_r(self):
        """HKL real axis names (read-only)."""
        return self._engine.axis_names_get(AXES_READ)

    @property
    def axes_w(self):
        """HKL real axis names (written by forward() computation)."""
        return self._engine.axis_names_get(AXES_WRITTEN)

    @property
    def mode(self):
        """HKL calculation mode (see also `HklCalc.modes`)"""
        return self._engine.current_mode_get()

    @mode.setter
    def mode(self, mode):
        if mode not in self.modes:
            raise ValueError("Unrecognized mode %r; choose from: %s" % (mode, ", ".join(self.modes)))

        return self._engine.current_mode_set(mode)

    @property
    def modes(self):
        return self._engine.modes_names_get()

    @property
    def solutions(self):
        """Allowed real solutions given the pseudo position."""
        return tuple(self._solutions)

    def update(self):
        """Calculate the pseudo axis positions from the real axis positions."""
        # TODO: though this works, maybe it could be named better on the hkl
        # side? either the 'get' function name or the fact that the EngineList
        # is more than just a list...

        self._engine_list.get()

    @property
    def parameters(self):
        """List the names of the additional parameters."""
        return self._engine.parameters_names_get()

    @property
    def parameters_values(self):
        """List the values of the additional parameters."""
        return self._engine.parameters_values_get(self._units)

    @property
    def pseudo_axis_names(self):
        """List the names of the pseudo axes."""
        return self._engine.pseudo_axis_names_get()

    @property
    def pseudo_axes(self):
        """List of pseudo axes as tuples: ``[(name, value)]``."""
        return OrderedDict(zip(self.pseudo_axis_names, self.pseudo_positions))

    @property
    def pseudo_positions(self):
        """List the values of the pseudo positioners."""
        return self._engine.pseudo_axis_values_get(self._units)

    @pseudo_positions.setter
    def pseudo_positions(self, values):
        try:
            geometry_list = self._engine.pseudo_axis_values_set(values, self._units)
        except util.GLib.GError as ex:
            raise ValueError("Calculation failed (%s)" % ex)

        Position = self._calc.Position

        def get_position(item):
            return Position(*item.geometry_get().axis_values_get(self._units))

        self._solutions = [get_position(item) for item in geometry_list.items()]

    def __getitem__(self, name):
        try:
            return self.pseudo_axes[name]
        except KeyError:
            raise ValueError("Unknown axis name: %s" % name)

    def __setitem__(self, name, value):
        values = self.pseudo_positions
        try:
            idx = self.pseudo_axis_names.index(name)
        except IndexError:
            raise ValueError("Unknown axis name: %s" % name)

        values[idx] = float(value)
        self.pseudo_positions = values

    @property
    def units(self):
        """The units used for calculations"""
        return self._calc.units

    @property
    def _units(self):
        """The (internal) units used for calculations"""
        return self._calc._units

    @property
    def engine(self):
        """The calculation engine"""
        return self._engine

    def _repr_info(self):
        r = [
            f"parameters={self.parameters!r}",
            f"pseudo_axes={self.pseudo_axes!r}",
            f"mode={self.mode!r}",
            f"modes={self.modes!r}",
            f"units={self.units!r}",
        ]

        return r

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._repr_info())})"


# when updating parameters we need to update the parent geometry object
class CalcParameter(Parameter):
    """
    Like calc parameter but needs reference to a geometry object.
    Updates to the parameter should be propagated back to the geometry.

    .. autosummary::

        ~fit
        ~limits
        ~value

    Parameters
    ----------
        param : HklParameter
        geometry: Geometry object
    """

    def __init__(self, param, geometry, *args, **kwargs):
        super().__init__(param, *args, **kwargs)
        self.param_name = param.name_get()
        # you may wonder why we stash geometry object rather than axis
        # this is because the axis params can only be updated from the geometry
        self._geometry = geometry

    @property
    def limits(self):
        """Allowed range of values (low, high)."""
        axis = self._geometry.axis_get(self.param_name)
        low, high = axis.min_max_get(self._units)
        if self._inverted:
            return (-high, -low)
        return low, high

    @limits.setter
    def limits(self, lims):
        low, high = lims
        if self._inverted:
            low, high = -high, -low
        self._param.min_max_set(low, high, self._units)
        # param name is the true name of axis
        axis = self._geometry.axis_get(self.param_name)
        axis.min_max_set(low, high, self._units)
        self._geometry.axis_set(self.param_name, axis)

    @property
    def value(self):
        """Position of this axis."""
        axis = self._geometry.axis_get(self.param_name)
        return axis.value_get(self._units)

    @value.setter
    def value(self, value):
        if self._inverted:
            value *= -1.0

        # param name is the true name of axis
        axis = self._geometry.axis_get(self.param_name)
        axis.value_set(value, self._units)
        self._geometry.axis_set(self.param_name, axis)

    @property
    def fit(self):
        """True if the parameter can be fit or not"""
        axis = self._geometry.axis_get(self.param_name)
        return bool(axis.fit_get())

    @fit.setter
    def fit(self, fit):
        # param name is the true name of axis
        axis = self._geometry.axis_get(self.param_name)
        axis.fit_set(fit)
        self._geometry.axis_set(self.param_name, axis)



"""
(Re)create the geometry_tables.rst document.

Run this manually, as needed::

    python \
        ./docs/make_geometry_tables.py \
        | tee ./docs/source/geometry_tables.rst

"""

import sys

import pyRestTable

from hkl import calc

PAGE_HEAD = f"""
.. this page created by {sys.argv[0]}

.. include:: /substitutions.txt

.. _geometry_tables:

=========================
Diffractometer Geometries
=========================

.. index:: mode

Tables are provided for the different geometries and then, for each geometry,
the calculation engines, pseudo axes required, modes of operation, and any
additional parameters required by the :ref:`mode <overview.mode>`.  The mode
defines which axes will be computed, which will be held constant, and any
relationships between axes.

Geometries indexed by number of circles
---------------------------------------

The different diffractometer geometries are distinguished, primarily, by
the number of axes (circles) and the names of each.  This table is
sorted first by the number of circles, and then the geometry name (as
used here in |hklpy|).

"""

TABLES_HEAD = """

.. _geometry_tables.tables:

Tables for each geometry
------------------------

A table is provided for each diffractometer geometry listing the calculation
engines, pseudo axes required, modes of operation, and any additional parameters
required by the mode.

* *engine* : Defines the names (and order) of the pseudo axes.
* *pseudo axes* : The engine performs
  :meth:`~hkl.diffract.Diffractometer.forward()` (pseudo-to-real) and
  :meth:`~hkl.diffract.Diffractometer.inverse()` (real-to-pseudo)
  transformations between the real-space axes and the *pseudo* (typically
  reciprocal-space) axes.  The *engine* defines the *pseudo axes* to be used.
* *mode* : Defines which axes are used for the ``forward()`` computation.
* *axes read* : Axes used in the ``forward()`` computation.
* *axes written* : Axes computed by the ``forward()`` computation.
* *extra parameters* : Any necessary additional parameters.
"""


def goniometers():
    for cname in dir(calc):
        if cname.startswith("Calc"):
            try:
                yield getattr(calc, cname)()
            except TypeError:
                pass


def print_summary_table():
    def format_name_list(names):
        names = [f"``{k}``" for k in names]
        return ", ".join(names)

    db = {
        gonio.geometry_name: {
            "real_axes": gonio.physical_axis_names,
            "cname": gonio.__class__.__name__[4:],  # 4 = len("Calc") as in "CalcE4CV"
        }
        for gonio in goniometers()
    }

    def sorter(gname):
        return f"{len(db[gname]['real_axes'])}-{gname}"

    table = pyRestTable.Table()
    table.addLabel("#circles")
    table.addLabel("geometry")
    table.addLabel("real_axes")
    for gname in sorted(db.keys(), key=sorter):
        gname_safe = gname.replace(" ", "_")
        entry = db[gname]
        real_axes = entry["real_axes"]
        table.addRow(
            [
                len(real_axes),
                f":ref:`{gname} <{gname_safe}_table>`",
                format_name_list(real_axes),
            ]
        )
    print(table)


def print_geometry_tables(rst=True):
    for gonio in goniometers():
        gonio.geometry_table(rst=rst)


if __name__ == "__main__":
    print(PAGE_HEAD.lstrip())
    print_summary_table()
    print(TABLES_HEAD)
    print_geometry_tables(rst=True)



"""
Utility functions and structures.

.. autosummary::

    ~Constraint
    ~diffractometer_types
    ~get_package_info
    ~get_position_tuple
    ~Lattice
    ~list_orientation_runs
    ~new_detector
    ~new_lattice
    ~restore_constraints
    ~restore_energy
    ~restore_orientation
    ~restore_reflections
    ~restore_sample
    ~restore_UB
    ~run_orientation_info
    ~software_versions

Also provides `SI_LATTICE_PARAMETER` as defined by the
*2018 CODATA recommended lattice parameter of silicon*. [#]_

.. [#] https://physics.nist.gov/cgi-bin/cuu/Value?asil
"""

import logging
import subprocess
import sys
from collections import defaultdict
from collections import namedtuple

import databroker
import gi
import numpy as np
import pandas as pd
import tqdm

gi.require_version("Hkl", "5.0")
from gi.repository import GLib  # noqa: F401
from gi.repository import Hkl as libhkl

__all__ = """
    Constraint
    diffractometer_types
    get_package_info
    get_position_tuple
    Lattice
    list_orientation_runs
    new_detector
    new_lattice
    restore_constraints
    restore_energy
    restore_orientation
    restore_reflections
    restore_sample
    restore_UB
    run_orientation_info
    SI_LATTICE_PARAMETER
    software_versions
""".split()
logger = logging.getLogger(__name__)


# when getting software package versions
DEFAULT_PACKAGE_LIST = "hkl pygobject".split()

# 2018 CODATA recommended lattice parameter of silicon, Angstrom.
# see: https://physics.nist.gov/cgi-bin/cuu/Value?asil
SI_LATTICE_PARAMETER = 5.431020511
# also provide the reported uncertainty, in case anyone is interested
SI_LATTICE_PARAMETER_UNCERTAINTY = 0.000000089


def new_detector(dtype=0):
    """Create a new HKL-library detector"""
    return libhkl.Detector.factory_new(libhkl.DetectorType(dtype))


if libhkl:
    diffractometer_types = tuple(sorted(libhkl.factories().keys()))
    UserUnits = libhkl.UnitEnum.USER
    DefaultUnits = libhkl.UnitEnum.DEFAULT

    # fmt: off
    units = {
        "user": UserUnits,
        "default": DefaultUnits
    }
    # fmt: on
else:
    diffractometer_types = ()
    units = {}


def to_numpy(mat):
    """Convert an hkl ``Matrix`` to a numpy ndarray

    Parameters
    ----------
    mat : Hkl.Matrix

    Returns
    -------
    ndarray
    """
    if isinstance(mat, np.ndarray):
        return mat

    ret = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            ret[i, j] = mat.get(i, j)

    return ret


def to_hkl(arr):
    """Convert a numpy ndarray to an hkl ``Matrix``

    Parameters
    ----------
    arr : ndarray

    Returns
    -------
    Hkl.Matrix
    """
    if isinstance(arr, libhkl.Matrix):
        return arr

    arr = np.array(arr)

    hklm = hkl_euler_matrix(0, 0, 0)
    hklm.init(*arr.flatten())
    return hklm


def hkl_euler_matrix(euler_x, euler_y, euler_z):
    """Convert into matrix form."""
    return libhkl.Matrix.new_euler(euler_x, euler_y, euler_z)


def _gi_info(gi_val):
    def get(attr):
        try:
            getter = getattr(gi_val, attr)
            # inspect.signature doesn't work on gi functions...
            return getter()
        except Exception as ex:
            try:
                return getter(units["user"])
            except Exception:
                return f"({ex.__class__.__name__}: {ex})"

    return {attr: get(attr) for attr in dir(gi_val) if attr.endswith("_get")}


class Constraint:
    """
    Limitations on acceptable positions from computed forward() solutions.

    Parameters
    ----------
    low_limit : float
        Lowest acceptable value for this axis when computing real-space solutions
        from given reciprocal-space positions.
    high_limit : float
        Highest acceptable value for this axis when computing real-space solutions
        from given reciprocal-space positions.
    value : float
        Constant value used (on condition) for ``forward(hkl)`` calculation.

        Implemented by diffractometer :attr:`~hkl.engine.Engine.mode`.

        The diffractometer engine's :attr:`~hkl.engine.Engine.mode` (such as E4CV's
        ``constant_phi`` mode) controls whether or not the axis is to be held
        constant.
    fit : bool
        (deprecated) Not used as a constraint.

        The value of ``fit`` is ignored.  It remains now for compatibility with
        previous *hklpy* releases. It will be dropped in a future *hklpy* release.

        While this parameter is used by *libhkl* to adjust lattice parameters when
        refining the **UB** matrix from more than 2 reflections, it is not used in
        the calculation of rotation angles from reciprocal-space coordinates.
    """

    def __init__(self, low_limit, high_limit, value, fit=True):
        self.low_limit = float(low_limit)
        self.high_limit = float(high_limit)
        self.value = float(value)
        self.fit = bool(fit)

        self._fields = "low_limit high_limit value fit".split()
        # fmt: off
        _fields = ", ".join(
            name + "=" + repr(getattr(self, name))
            for name in self._fields
        )
        # fmt: on
        self._repr_fmt = f"({_fields})"

    def _asdict(self):
        "Return a new dict which maps field names to their values."
        return dict(zip(self._fields, self))

    class _ConstraintIterator:
        "Iterator"

        def __init__(self, constraint):
            self.constraint = constraint
            self._index = 0

        def __next__(self):
            if self._index < len(self.constraint._fields):
                c = getattr(self.constraint, self.constraint._fields[self._index])
                self._index += 1
                return c
            else:
                raise StopIteration

    def __iter__(self):
        "Iterate through the fields."
        return self._ConstraintIterator(self)

    def __repr__(self):
        "Return a nicely formatted representation string."
        content = "(" + ", ".join([f"{k}={getattr(self, k)}" for k in self._fields]) + ")"
        return self.__class__.__name__ + content


Lattice = namedtuple("LatticeTuple", "a b c alpha beta gamma")


def new_lattice(a, b=None, c=None, alpha=90.0, beta=None, gamma=None):
    """
    Simplify for high-symmetry crystal systems.

    EXAMPLES (highest to lowest symmetry):

    =============== =================================== === === === ======= ======= =====
    system          command                             a   b   c   alpha   beta    gamma
    =============== =================================== === === === ======= ======= =====
    cubic           new_lattice(5.)                     5   5   5   90      90      90
    hexagonal       new_lattice(4., c=3., gamma=120)    4   4   3   90      90      120
    rhombohedral    new_lattice(4., alpha=80.0)         4   4   4   80      80      80
    tetragonal      new_lattice(4, c=3)                 4   4   3   90      90      90
    orthorhombic    new_lattice(4, 5, 3)                4   5   3   90      90      90
    monoclinic      new_lattice(4, 5, 3, beta=75)       4   5   3   90      75      90
    triclinic       new_lattice(4, 5, 3, 75., 85., 95.) 4   5   3   75      85      95
    =============== =================================== === === === ======= ======= =====

    .. see: https://en.wikipedia.org/wiki/Crystal_system
    """
    return Lattice(a, b or a, c or a, alpha, beta or alpha, gamma or alpha)


_position_tuples = {}


def get_position_tuple(axis_names, class_name="Position"):
    """Return a namedtuple with the positions."""
    global _position_tuples

    key = frozenset(axis_names)
    if key not in _position_tuples:
        _position_tuples[key] = namedtuple(class_name, tuple(axis_names))

    return _position_tuples[key]


def list_orientation_runs(catalog, *args, limit=20):
    """
    List the runs with orientation information.

    Returns
    -------
    orientation information: Pandas DataFrame object

    Parameters
    ----------
    *args : [str]
        A list of additional data column names to be displayed,
        corresponding to the names of available orientation
        information in the descriptor document.

        Example::

            list_orientation_runs(
                "class_name", energy", "energy_units", "lattice")

    catalog : object
        Instance of a databroker catalog.
    limit : int
        Limit the list to the first ``limit`` rows. (default=20)
    """
    buffer = []
    _count = 0
    default_columns = "sample_name diffractometer_name geometry_name".split()
    display_columns = default_columns + list(args)
    with tqdm.tqdm(total=len(catalog.v2), file=sys.stdout) as progress_bar:
        for run in catalog.v2.values():
            info = run_orientation_info(run)
            if len(info):
                scan_id = run.metadata["start"]["scan_id"]
                if databroker.__version__ >= "2.0":
                    uid = run.start["uid"][:7]
                else:
                    uid = run.name[:7]
                for device in sorted(info.keys()):
                    orientation = info[device]
                    row = dict(scan_id=scan_id)
                    for f in display_columns:
                        if f in orientation:
                            row[f] = orientation[f]
                    row["uid"] = uid
                    buffer.append(row)
                    _count += 1
                    if _count >= limit:
                        break
            if _count >= limit:
                break
            progress_bar.update()
    return pd.DataFrame(buffer)


def run_orientation_info(run):
    """
    Return a dictionary with orientation information in this run.

    Dictionary is keyed by "detector" name (in case more than one
    diffractometer was added as a "detector" to the run).

    The orientation information is found in the descriptor document
    of the primary stream.

    Parameters
    ----------
    run : from Databroker
        A Bluesky run, from databroker v2, such as ``cat.v2[-1]``.
    """
    devices = {}

    try:
        if databroker.__version__ >= "2.0":
            for descriptor in run.primary.descriptors:
                for device, configuration in descriptor.get("configuration", {}).items():
                    conf = configuration.get("data", {})
                    if f"{device}_orientation_attrs" in conf:
                        # fmt:off
                        devices[device] = {
                            item[len(f"{device}_"):]: value
                            for item, value in conf.items()
                        }
                        # fmt:on

        else:  # older databroker v1.2
            run_conf = run.primary.config
            for device in sorted(run_conf):
                conf = run_conf[device].read()
                if f"{device}_orientation_attrs" in conf:
                    # fmt:off
                    devices[device] = {
                        item[len(f"{device}_"):]: conf[item].to_dict()["data"][0]
                        for item in conf
                    }
                    # fmt:on
    except Exception as exc:
        logger.warning("Could not process run %s, due to %s", run, exc)
    return devices


def _smart_signal_update(value, signal):
    """Write value to signal if not equal.  Not a plan."""
    if signal.get() != value:
        signal.put(value)


def _check_geometry(orientation, diffractometer):
    """
    Check that geometry of recovered orientation matches diffractometer.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.

    Raise ValueError if mismatched.
    """
    if orientation["geometry_name"] != diffractometer.geometry_name.get():
        raise ValueError(
            "Geometries do not match:"
            f" Orientation={orientation['geometry_name']},"
            f" {diffractometer.name}={diffractometer.geometry_name.get()},"
            " will not restore."
        )


def restore_constraints(orientation, diffractometer):
    """
    Restore any constraints from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)

    # fmt:off
    if len(orientation["_constraints"]):
        con_dict = {
            k: Constraint(*v)
            for k, v in zip(orientation["_reals"], orientation["_constraints"])
        }
        diffractometer.apply_constraints(con_dict)
    # fmt:on


def restore_energy(orientation, diffractometer):
    """
    Restore energy from orientation information.

    NOTE: This makes blocking calls so do not use in bluesky plan.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    # get _all_ the expected keys
    try:
        kv_dict = {
            orientation[attr]: getattr(diffractometer, attr)
            for attr in "energy energy_units energy_offset".split()
        }
    except KeyError as exc:
        # fmt: off
        raise KeyError(
            f"{diffractometer.name}: Cannot restore "
            f"diffractometer energy due to missing {exc} term."
        )
        # fmt: on
    # update the signals
    for k, v in kv_dict.items():
        _smart_signal_update(k, v)


def restore_reflections(orientation, diffractometer):
    """
    Restore any reflections from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)
    calc = diffractometer.calc
    # remember this wavelength
    wavelength0 = calc.wavelength

    # short aliases
    pseudos = orientation["_pseudos"]
    reals = orientation["_reals"]
    if reals != calc.physical_axis_names and reals == calc._geometry.axis_names_get():
        # Substitute user-defined axes names for canonical axes names.
        reals = calc.physical_axis_names
    orientation_reflections = []
    # might be renamed axes
    renaming = calc._axis_name_to_original

    for ref_base in orientation["reflections_details"]:
        # every reflection has its own wavelength
        calc.wavelength = ref_base["wavelength"]

        # Order of the items is important.
        # Can't just use the dictionaries in ``orientation``.
        # Get the canonical order from the orientation data.
        miller_indices = [ref_base["reflection"][key] for key in pseudos]
        try:
            positions = [ref_base["position"][key] for key in reals]
        except KeyError:
            # switch to use renamed keys
            positions = [ref_base["position"][renaming[key]] for key in reals]
        ppp = namedtuple("PositionTuple", tuple(reals))(*positions)

        # assemble the final form of the reflection for add_reflection()
        reflection = tuple([*miller_indices, ppp])
        r = calc.sample.add_reflection(*reflection)
        if ref_base["orientation_reflection"]:
            orientation_reflections.append(r)

    if len(orientation_reflections) > 1:
        # compute **UB** from the last two orientation reflections
        calc.sample.compute_UB(*orientation_reflections[-2:])

    # restore previous wavelength
    if calc.wavelength != wavelength0:
        calc.wavelength = wavelength0


def restore_orientation(orientation, diffractometer):
    """
    Restore all orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)
    restore_energy(orientation, diffractometer)
    restore_sample(orientation, diffractometer)
    restore_constraints(orientation, diffractometer)
    restore_reflections(orientation, diffractometer)


def restore_sample(orientation, diffractometer):
    """
    Restore sample & lattice from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    nm = orientation["sample_name"]
    lattice = orientation["lattice"]
    if nm not in diffractometer.calc._samples:
        diffractometer.calc.new_sample(nm, lattice=lattice)
    else:
        raise ValueError(f"Sample '{nm}' already exists in {diffractometer.name}.")


def restore_UB(orientation, diffractometer):
    """
    Restore **UB** matrix from orientation information.

    Parameters
    ----------
    orientation : dict
        Dictionary of orientation parameters (from
        :func:`~hkl.util.run_orientation_info()`) recovered from run.
    diffractometer : :class:`~hkl.diffract.Diffractometer()`
        Diffractometer object.
    """
    _check_geometry(orientation, diffractometer)
    _smart_signal_update(orientation["UB"], diffractometer.UB)


def _installed_package_information():
    """Index information about packages known by conda and/or pip."""
    packages = defaultdict(dict)

    for tool in "conda pip".split():
        s = subprocess.run([tool, "list"], capture_output=True)
        for line in s.stdout.splitlines():
            if not line.decode().startswith("#"):
                args = line.decode().strip().split()
                key = args[0]
                packages[key]["version"] = args[1]
                packages[key][tool] = True
                if tool == "conda":
                    packages[key]["build"] = args[2]
                    packages[key]["conda"] = True
                    if len(args) == 4:
                        packages[key]["channel"] = args[3]

                elif tool == "pip":
                    if len(args) == 3:
                        packages[key]["location"] = args[2]

    return packages


# cache of discovered installed package information
_package_info = None


def get_package_info(package_name):
    """Return dict of information about installed package, by name."""
    global _package_info
    if _package_info is None:
        # index the known packages
        # This is not expected to change once the process has started.
        _package_info = _installed_package_information()
    return _package_info.get(package_name)


def software_versions(keys=[]):
    """
    Report the package versions, in a dictionary.

    EXAMPLE::

        In [1]: import gi
        ...:
        ...: gi.require_version("Hkl", "5.0")
        ...:
        ...: import hkl.util

        In [2]: hkl.util.software_versions()
        Out[2]:
        {'hkl': '5.0.0.2173',
        'hklpy': '0.3.16+131.ga5a449a.dirty',
        'pygobject': '3.40.1'}

    Here, it shows (albeit indirectly) *Hkl 5.0.0 tag 2173*.  Proceed to the Hkl
    source repository, list of tags [#]_, and find tag ``2173`` [#]_.

    .. [#] Hkl source tags: https://repo.or.cz/hkl.git/refs
    .. [#] Hkl tag 2173: https://repo.or.cz/hkl.git/tree/refs/tags/v5.0.0.2173)
    """
    if keys is None or len(keys) == 0:
        keys = DEFAULT_PACKAGE_LIST
    v_dict = {}
    for key in keys:
        info = get_package_info(key)
        if info is not None:
            v_dict[key] = info.get("version")
    return v_dict



"""
:mod:`hkl` - HKL diffractometer support
=======================================

.. module:: hkl
   :synopsis:

"""

# flake8: noqa

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

PROJECT_NAME = "hklpy"  # pyproject.toml::[project]name

try:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)
    del get_version
except (LookupError, ModuleNotFoundError):
    from importlib.metadata import version

    __version__ = version(PROJECT_NAME)
    del version

# gobject-introspection, to access libhkl
import gi

gi.require_version("Hkl", "5.0")

# import shortcuts

from .calc import A_KEV  # noqa: F401, F402, E402
from .calc import UnreachableError  # noqa: F401, F402, E402
from .configuration import DiffractometerConfiguration  # noqa: F401, F402, E402
from .geometries import E4CH  # noqa: F401, F402, E402
from .geometries import E4CV
from .geometries import E6C
from .geometries import K4CV
from .geometries import K6C
from .geometries import ApsPolar
from .geometries import Petra3_p09_eh2
from .geometries import Petra3_p23_4c
from .geometries import Petra3_p23_6c
from .geometries import SimMixin
from .geometries import SimulatedE4CV
from .geometries import SimulatedE6C
from .geometries import SimulatedK4CV
from .geometries import SimulatedK6C
from .geometries import SoleilMars
from .geometries import SoleilNanoscopiumRobot
from .geometries import SoleilSiriusKappa
from .geometries import SoleilSiriusTurret
from .geometries import SoleilSixsMed1p2
from .geometries import SoleilSixsMed2p2
from .geometries import SoleilSixsMed2p3
from .geometries import SoleilSixsMed2p3v2
from .geometries import Zaxis
from .user import cahkl  # noqa: F401, F402, E402
from .user import cahkl_table
from .user import calc_UB
from .user import change_sample
from .user import list_samples
from .user import new_sample
from .user import or_swap
from .user import pa
from .user import select_diffractometer
from .user import set_energy
from .user import setor
from .user import show_sample
from .user import show_selected_diffractometer
from .user import update_sample
from .user import wh
from .util import SI_LATTICE_PARAMETER  # noqa: F401, F402, E402
from .util import Constraint
from .util import Lattice
from .util import diffractometer_types
from .util import get_position_tuple
from .util import list_orientation_runs
from .util import new_detector
from .util import new_lattice
from .util import restore_constraints
from .util import restore_energy
from .util import restore_orientation
from .util import restore_reflections
from .util import restore_sample
from .util import restore_UB
from .util import run_orientation_info
from .util import software_versions



import pathlib
from contextlib import nullcontext as does_not_raise
from dataclasses import MISSING

import pytest
from apischema import ValidationError
from apischema import deserialize

from .. import DiffractometerConfiguration
from ..configuration import EXPORT_FORMATS
from ..configuration import DCConfiguration
from ..configuration import DCConstraint
from ..configuration import DCLattice
from ..configuration import DCReflection
from ..configuration import DCSample
from ..util import Constraint
from ..util import new_lattice
from .tools import TWO_PI

TEST_CONFIG_FILE = "data/e4c-config.json"


def test_e4cv(e4cv):
    assert e4cv is not None


def test_axes_names(e4cv, e4cv_renamed):
    c_e4cv = DiffractometerConfiguration(e4cv)
    c_e4cv_renamed = DiffractometerConfiguration(e4cv_renamed)
    assert c_e4cv.canonical_axes_names == c_e4cv_renamed.canonical_axes_names
    assert c_e4cv.real_axes_names != c_e4cv_renamed.real_axes_names
    assert c_e4cv.reciprocal_axes_names == c_e4cv_renamed.reciprocal_axes_names


def test_restore(e4cv_renamed, k4cv):
    config = DiffractometerConfiguration(e4cv_renamed)
    before = config.export("dict")
    config.restore(before)
    after = config.export("dict")
    assert before != after, "datetime difference"

    before.pop("datetime")
    after.pop("datetime")
    assert before == after, "should be same configuration"

    with pytest.raises(ValueError):
        # cannot restore k4cv config to e4cv diffractometer
        config.restore(DiffractometerConfiguration(k4cv).export())
    after = config.export("dict")
    after.pop("datetime")
    assert before == after, "configuration should be unchanged"


@pytest.mark.parametrize("fmt", [None] + EXPORT_FORMATS)
def test_format(fmt, e4cv):
    """Verify that various export formats can be imported."""
    config = DiffractometerConfiguration(e4cv)
    with does_not_raise():
        cfg = config.export(fmt)
    assert isinstance(cfg, (dict, str)), f"{cfg=}"

    with does_not_raise():
        # first, test restore by known type
        if fmt in [None, "json"]:
            config.from_json(cfg)  # default format
        elif fmt == "dict":
            config.from_dict(cfg)
        elif fmt == "yaml":
            config.from_yaml(cfg)
        else:
            raise ValueError(f"Unrecognized configuration format: {fmt}")
        config.restore(cfg)  # test restore with automatic type recognition


@pytest.mark.parametrize("file", [None, TEST_CONFIG_FILE])  # default or restored config
@pytest.mark.parametrize("action", "rm set".split())  # remove or set keys incorrectly
@pytest.mark.parametrize(
    "key, value, failure",  # fmt: off
    [
        [k, object, (KeyError, TypeError, ValidationError, ValueError)]
        for k in DCConfiguration.__dataclass_fields__
    ],  # fmt: off
)
def test_validation_fails(file, action, key, value, failure, e4cv):
    agent = DiffractometerConfiguration(e4cv)
    assert isinstance(agent, DiffractometerConfiguration), f"{agent}"

    if file is not None:
        path = pathlib.Path(__file__).parent / file
        assert path.exists(), f"{path}"
        agent.restore(path)

    data = agent.export("dict")
    assert isinstance(data, dict), f"{type(data)=}"

    if action == "rm":
        data.pop(key)
        # Determine if key is optional (default/factory is defined).  Empirical.
        attr = DCConfiguration.__dataclass_fields__[key]
        if attr.default_factory != MISSING or attr.default != MISSING:
            failure = None  # OK if not provided
    elif action == "set":
        data[key] = value

    if failure is None:
        agent.restore(data)
    else:
        with pytest.raises(failure):
            agent.restore(data)


def common_DC_dataclass_tests(dc_class, data, key, value, failure, val_arg):
    """Code used to test the DCxyz classes."""
    agent = deserialize(dc_class, data)
    assert isinstance(agent, dc_class)
    assert key in data

    # Is key an optional attribute?
    context = pytest.raises(ValidationError)
    attr = dc_class.__dataclass_fields__[key]
    if attr.default_factory != MISSING or attr.default != MISSING:
        context = does_not_raise()

    data.pop(key)
    with context:
        deserialize(dc_class, data)

    data[key] = value
    with failure:
        agent = deserialize(dc_class, data)
        assert isinstance(agent, dc_class)
        agent.validate(val_arg)


@pytest.mark.parametrize("key", "low_limit high_limit value".split())
@pytest.mark.parametrize(
    "value, failure",
    [
        [-360.01, pytest.raises(ValueError)],
        [-360, does_not_raise()],
        [0, does_not_raise()],
        [360, does_not_raise()],
        [360.01, pytest.raises(ValueError)],
        ["0", pytest.raises(ValidationError)],
    ],
)
def test_DCConstraint_fails(key, value, failure):
    # all attributes are required
    # fmt: off
    data = {
        "low_limit": 0.0,
        "high_limit": 0.0,
        "value": 0.0,
        "fit": True,
    }
    common_DC_dataclass_tests(
        DCConstraint, data, key, value, failure, f"testing DCConstraint.{key=}"
    )
    # fmt: on


@pytest.mark.parametrize("key", "a b c alpha beta gamma".split())
@pytest.mark.parametrize(
    "value, failure",
    [
        [-1, pytest.raises(ValueError)],
        [0, pytest.raises(ValueError)],
        [1e-7, pytest.raises(ValueError)],
        [1, does_not_raise()],
        [179.99, does_not_raise()],
        [180, does_not_raise()],
        [10_000, does_not_raise()],
        [100_000_000, pytest.raises(ValueError)],
    ],
)
def test_DCLattice_fails(key, value, failure, e4cv):
    data = {
        "a": 4,
        "b": 5,
        "c": 6,
        "alpha": 8,
        "beta": 9,
        "gamma": 10,
    }
    if key in "alpha beta gamma".split() and value > 180 - 1e-6:
        failure = pytest.raises(ValueError)

    agent = DiffractometerConfiguration(e4cv)
    common_DC_dataclass_tests(DCLattice, data, key, value, failure, agent)


@pytest.mark.parametrize(
    "key, value, failure",
    [
        ["wavelength", -1, pytest.raises(ValueError)],
        ["wavelength", 0, pytest.raises(ValueError)],
        ["wavelength", 0.01, does_not_raise()],
        ["wavelength", 500, does_not_raise()],
        ["wavelength", 1_000_000.0, does_not_raise()],
        ["wavelength", 100_000_000, pytest.raises(ValueError)],
        ["reflection", {"h": 12.4, "k": 0, "l": 0}, does_not_raise()],
        ["reflection", {"h": 0, "k": 12.4, "l": 0}, does_not_raise()],
        ["reflection", {"h": 0, "k": 0, "l": 12.4}, does_not_raise()],
        ["reflection", {"h": -12.4, "k": 0, "l": 0}, does_not_raise()],
        ["reflection", {"h": 0, "k": -12.4, "l": 0}, does_not_raise()],
        ["reflection", {"h": 0, "k": 0, "l": -12.4}, does_not_raise()],
        ["reflection", {"h": 20_000, "k": 0, "l": 0}, pytest.raises(ValueError)],
        ["reflection", {"h": -20_000, "k": 0, "l": 0}, pytest.raises(ValueError)],
        ["reflection", {"h": 0, "k": 20_000, "l": 0}, pytest.raises(ValueError)],
        ["reflection", {"h": 0, "k": -20_000, "l": 0}, pytest.raises(ValueError)],
        ["reflection", {"h": 0, "k": 0, "l": 20_000}, pytest.raises(ValueError)],
        ["reflection", {"h": 0, "k": 0, "l": -20_000}, pytest.raises(ValueError)],
        ["position", {"omega": 360, "chi": 0, "phi": 0, "tth": 0}, does_not_raise()],
        ["position", {"omega": 0, "chi": 360, "phi": 0, "tth": 0}, does_not_raise()],
        ["position", {"omega": 0, "chi": 0, "phi": 360, "tth": 0}, does_not_raise()],
        ["position", {"omega": 0, "chi": 0, "phi": 0, "tth": 360}, does_not_raise()],
        ["position", {"omega": -360, "chi": 0, "phi": 0, "tth": 0}, does_not_raise()],
        ["position", {"omega": 0, "chi": -360, "phi": 0, "tth": 0}, does_not_raise()],
        ["position", {"omega": 0, "chi": 0, "phi": -360, "tth": 0}, does_not_raise()],
        ["position", {"omega": 0, "chi": 0, "phi": 0, "tth": -360}, does_not_raise()],
        ["position", {"omega": 360.01, "chi": 0, "phi": 0, "tth": 0}, pytest.raises(ValueError)],
        ["position", {"omega": 0, "chi": 360.01, "phi": 0, "tth": 0}, pytest.raises(ValueError)],
        ["position", {"omega": 0, "chi": 0, "phi": 360.01, "tth": 0}, pytest.raises(ValueError)],
        ["position", {"omega": 0, "chi": 0, "phi": 0, "tth": 360.01}, pytest.raises(ValueError)],
        ["position", {"omega": -360.01, "chi": 0, "phi": 0, "tth": 0}, pytest.raises(ValueError)],
        ["position", {"omega": 0, "chi": -360.01, "phi": 0, "tth": 0}, pytest.raises(ValueError)],
        ["position", {"omega": 0, "chi": 0, "phi": -360.01, "tth": 0}, pytest.raises(ValueError)],
        ["position", {"omega": 0, "chi": 0, "phi": 0, "tth": -360.01}, pytest.raises(ValueError)],
    ],
)
def test_DCReflection_fails(key, value, failure, e4cv):
    data = {
        "reflection": {"h": 0, "k": 0, "l": 0},
        "position": {"omega": 0, "chi": 0, "phi": 0, "tth": 0},
        "wavelength": 1,
        "orientation_reflection": True,
    }
    agent = DiffractometerConfiguration(e4cv)
    common_DC_dataclass_tests(DCReflection, data, key, value, failure, agent)


@pytest.mark.parametrize(
    "key, value, failure",
    [
        ["name", "", pytest.raises(ValueError)],
        ["name", "   ", pytest.raises(ValueError)],
        ["lattice", object, pytest.raises(TypeError)],
        ["lattice", {}, pytest.raises(ValidationError)],
        ["lattice", {"error", object}, pytest.raises(TypeError)],
        ["reflections", [object], pytest.raises(TypeError)],
        ["U", [object], pytest.raises(TypeError)],
        ["UB", [object], pytest.raises(TypeError)],
        ["U", [[1, 0, object], [0, 0, 0], [0, 0, 0]], pytest.raises(TypeError)],
        ["UB", [[1, 0, object], [0, 0, 0], [0, 0, 0]], pytest.raises(TypeError)],
        ["U", [[1, 0, 0], [0, 0, 0]], pytest.raises(ValueError)],
        ["UB", [[1, 0, 0], [0, 0, 0]], pytest.raises(ValueError)],
    ],
)
def test_DCSample_fails(key, value, failure, e4cv):
    data = {
        "name": "vibranium",
        "lattice": {"a": 4, "b": 4, "c": 4, "alpha": 90, "beta": 90, "gamma": 90},
        "reflections": [],
        "U": [[1.0, 0.0, 0.0], [0.0, 1.0, 1.0], [0.0, 0.0, 1.0]],
        "UB": [[TWO_PI, 0.0, 0.0], [0.0, TWO_PI, 0.0], [0.0, 0.0, TWO_PI]],
    }
    agent = DiffractometerConfiguration(e4cv)
    common_DC_dataclass_tests(DCSample, data, key, value, failure, agent)


@pytest.mark.parametrize("clear", [True, False, object, None])
def test_diffractometer_restored(clear, e4cv):
    # -------------------------------- default configuration
    mode_default = e4cv.engine.mode
    positions_before = e4cv.RealPosition
    reciprocal_positions_before = e4cv.PseudoPosition
    constraints_default = e4cv._constraints_for_databroker

    assert round(e4cv.calc.wavelength, 2) == 1.55
    assert e4cv.engine.mode == mode_default
    assert len(e4cv.calc._samples) == 1
    assert e4cv.calc.sample.name == "main"
    assert len(e4cv.calc.sample.reflections_details) == 0

    # -------------------------------- perturb from the defaults
    mode_changed = e4cv.engine.modes[-1]
    constraints_test = {
        "omega": Constraint(-10, 80, 10, 1),
        "chi": Constraint(-10, 91, 45, 0),
        "phi": Constraint(-10, 91, 90, 0),
        "tth": Constraint(-10, 120, 20, 1),
    }
    wavelength_test = 0.25

    e4cv.engine.mode = mode_changed
    assert e4cv.engine.mode == mode_changed

    assert e4cv._constraints_for_databroker == constraints_default
    assert e4cv.calc.wavelength != wavelength_test

    e4cv.apply_constraints(constraints_test)
    constraints_changed = e4cv._constraints_for_databroker
    assert constraints_changed != constraints_default

    e4cv.calc.wavelength = wavelength_test
    assert e4cv.calc.wavelength == wavelength_test

    # modify the sample lattice
    orthorhombic = new_lattice(4, 5, 6)
    e4cv.calc.sample.lattice = orthorhombic
    assert e4cv.calc.sample.lattice == orthorhombic

    # add a reflection
    assert len(e4cv.calc.sample.reflections_details) == 0
    e4cv.calc.sample.add_reflection(0.1, 0, 0, (1, 2, 3, 4))
    assert len(e4cv.calc.sample.reflections_details) == 1

    # -------------------------------- restore config from file
    test_file = pathlib.Path(__file__).parent / TEST_CONFIG_FILE
    assert test_file.exists()

    agent = DiffractometerConfiguration(e4cv)
    full_config_before = agent.export("dict")
    context = does_not_raise()
    if not isinstance(clear, bool):
        context = pytest.raises(TypeError)
    with context:
        agent.restore(test_file, clear=clear)

    # -------------------------------- these should not have been changed
    assert e4cv.calc.wavelength == wavelength_test, "wavelength changed"
    assert e4cv.RealPosition == positions_before, "RealPosition changed"
    assert e4cv.PseudoPosition == reciprocal_positions_before, "PseudoPosition changed"

    # -------------------------------- changes conditional on clear

    full_config_after = agent.export("dict")
    full_config_before.pop("datetime")
    full_config_after.pop("datetime")

    if isinstance(clear, bool):
        assert full_config_after != full_config_before
        assert e4cv._constraints_for_databroker != constraints_default
        assert e4cv._constraints_for_databroker != constraints_changed
        # one additional sample
        assert len(e4cv.calc._samples) == 2
        assert list(e4cv.calc._samples) == "main vibranium".split()

        if clear:
            assert e4cv.engine.mode == mode_default
            assert e4cv.engine.mode != mode_changed
            assert e4cv.calc._samples["main"].lattice != orthorhombic
            assert len(e4cv.calc._samples["main"].reflections_details) == 0
        else:
            # switch samples
            e4cv.calc.sample = "vibranium"
            assert len(e4cv.calc.sample.reflections_details) == 3
            e4cv.calc.sample.lattice.a == TWO_PI
            e4cv.calc.sample.lattice.b == TWO_PI
            e4cv.calc.sample.lattice.c == TWO_PI
    else:
        # when restore() failed
        assert full_config_after == full_config_before
        assert len(e4cv.calc._samples) == 1
        assert list(e4cv.calc._samples) == "main".split()
        assert e4cv.engine.mode != mode_default
        assert e4cv.engine.mode == mode_changed
        assert e4cv.calc._samples["main"].lattice == orthorhombic
        assert len(e4cv.calc._samples["main"].reflections_details) == 1


@pytest.mark.parametrize("restore_constraints", [True, False, object, None])
def test_diffractometer_constraints_restored(restore_constraints, e4cv):
    """Are the constraints restored by option value?"""
    test_file = pathlib.Path(__file__).parent / TEST_CONFIG_FILE
    assert test_file.exists()

    agent = DiffractometerConfiguration(e4cv)
    constraints_before = agent.export("dict")["constraints"]

    context = does_not_raise()
    if not isinstance(restore_constraints, bool):
        context = pytest.raises(TypeError)
    with context:
        agent.restore(test_file, restore_constraints=restore_constraints)

    constraints_now = agent.export("dict")["constraints"]
    if isinstance(restore_constraints, bool) and restore_constraints:
        assert constraints_before != constraints_now
    else:
        assert constraints_before == constraints_now


def test_export_to_file(e4cv, tmp_path):
    """Can configuration be exported to file using pathlib object?"""
    assert isinstance(tmp_path, pathlib.Path)
    assert tmp_path.exists()

    path = tmp_path / "config.txt"
    assert not path.exists()

    agent = DiffractometerConfiguration(e4cv)
    agent.export(path)
    assert path.exists()

    agent.restore(path)  # just to be safe here


def test_constraints_stack(e4cv):
    """Ensure that restored constraints can be removed by undo."""
    agent = DiffractometerConfiguration(e4cv)
    constraints_before = agent.export("dict")["constraints"]

    test_file = pathlib.Path(__file__).parent / TEST_CONFIG_FILE
    assert test_file.exists()

    agent.restore(test_file)
    assert agent.export("dict")["constraints"] != constraints_before

    e4cv.undo_last_constraints()
    assert agent.export("dict")["constraints"] == constraints_before


def test_preview(e4cv):
    agent = DiffractometerConfiguration(e4cv)

    report = agent.preview(agent.export()).splitlines()
    assert report[0] == f"name: {e4cv.name}"
    assert report[2] == f"geometry: {e4cv.geometry_name.get()}"
    assert report[4] == "Table of Samples"
    assert "====" in report[5]
    assert "sample" in report[6]
    assert "main" in report[8]
    assert len(report) == 10

    report = agent.preview(agent.export(), show_reflections=True).splitlines()
    assert len(report) == 10  # no reflections, no new info

    report = agent.preview(agent.export(), show_constraints=True).splitlines()
    assert len(report) == 10 + 11  # new lines
    assert report[12] == "Table of Axis Constraints"
    assert "====" in report[13]
    assert "fit?" in report[14]
    assert report[16].startswith("omega")
    assert report[19].startswith("tth")

    main = e4cv.calc.sample  # the default sample
    m_100 = main.add_reflection(1, 0, 0, (-45, 0, 0, 0))
    m_010 = main.add_reflection(0, 1, 0, (45, 0, 0, 0))
    main.add_reflection(0, 0, 1, (45, 0, 90, 0))
    main.compute_UB(m_100, m_010)
    report = agent.preview(agent.export(), show_reflections=True).splitlines()
    assert len(report) == 10 + 10  # new lines
    assert report[12] == f"Table of Reflections for Sample: {main.name}"
    assert report[14].split()[1] == "h"
    assert report[14].split()[-3] == "tth"
    assert report[14].split()[-2] == "wavelength"
    assert report[14].split()[-1] == "orient?"
    assert report[16].split()[1] == "1"
    assert report[16].split()[2] == "0"
    assert report[16].split()[3] == "0"
    assert report[16].split()[-1] == "True"
    assert report[17].split()[1] == "0"
    assert report[17].split()[2] == "1"
    assert report[17].split()[3] == "0"
    assert report[17].split()[-1] == "True"
    assert report[18].split()[1] == "0"
    assert report[18].split()[2] == "0"
    assert report[18].split()[3] == "1"
    assert report[18].split()[-1] == "False"



"""
diffract
--------

Support for diffractometer instances.

SUPPORT

.. autosummary::

    ~SimMixin

DIFFRACTOMETER GEOMETRIES

.. autosummary::

    ~E4CH
    ~E4CV
    ~E6C
    ~K4CV
    ~K6C
    ~Zaxis
    ~SimulatedE4CV
    ~SimulatedE6C
    ~SimulatedK4CV
    ~SimulatedK6C

SPECIAL-USE DIFFRACTOMETER GEOMETRIES

.. autosummary::

    ~ApsPolar
    ~Petra3_p09_eh2
    ~Petra3_p23_4c
    ~Petra3_p23_6c
    ~SoleilMars
    ~SoleilNanoscopiumRobot
    ~SoleilSiriusKappa
    ~SoleilSiriusTurret
    ~SoleilSixsMed1p2
    ~SoleilSixsMed2p2
    ~SoleilSixsMed2p3
    ~SoleilSixsMed2p3v2

"""

import logging

from ophyd import Component as Cpt
from ophyd import Device
from ophyd import PseudoSingle
from ophyd import SoftPositioner

from . import calc
from .diffract import Diffractometer

__all__ = """
    ApsPolar
    E4CH
    E4CV
    E6C
    K4CV
    K6C
    Petra3_p09_eh2
    Petra3_p23_4c
    Petra3_p23_6c
    SimMixin
    SimulatedE4CV
    SimulatedE6C
    SimulatedK4CV
    SimulatedK6C
    SoleilMars
    SoleilNanoscopiumRobot
    SoleilSiriusKappa
    SoleilSiriusTurret
    SoleilSixsMed1p2
    SoleilSixsMed2p2
    SoleilSixsMed2p3
    SoleilSixsMed2p3v2
    Zaxis
""".split()
logger = logging.getLogger(__name__)


class ApsPolar(Diffractometer):
    """APS POLAR 6-circle diffractometer."""

    calc_class = calc.CalcApsPolar


class E4CH(Diffractometer):
    """Eulerian 4-circle, horizontal scattering plane"""

    calc_class = calc.CalcE4CH


class E4CV(Diffractometer):
    """Eulerian 4-circle, vertical scattering plane"""

    calc_class = calc.CalcE4CV


class E6C(Diffractometer):
    """Eulerian 6-circle, vertical scattering plane"""

    calc_class = calc.CalcE6C


class K4CV(Diffractometer):
    """Kappa 4-circle, vertical scattering plane"""

    calc_class = calc.CalcK4CV


class K6C(Diffractometer):
    """Kappa 6-circle, vertical scattering plane"""

    calc_class = calc.CalcK6C


class Petra3_p09_eh2(Diffractometer):
    """6-circle Used at Petra3 P09"""

    calc_class = calc.CalcPetra3_p09_eh2


class Petra3_p23_4c(Diffractometer):
    """4-circle Used at Petra3 P23"""

    calc_class = calc.CalcPetra3_p23_4c


class Petra3_p23_6c(Diffractometer):
    """7-circle Used at Petra3 P23"""

    calc_class = calc.CalcPetra3_p23_6c


class SoleilMars(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilMars


class SoleilNanoscopiumRobot(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilNanoscopiumRobot


class SoleilSiriusKappa(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilSiriusKappa


class SoleilSiriusTurret(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilSiriusTurret


class SoleilSixsMed1p2(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilSixsMed1p2


class SoleilSixsMed2p2(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilSixsMed2p2


class SoleilSixsMed2p3(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilSixsMed2p3


class SoleilSixsMed2p3v2(Diffractometer):
    """Used at Soleil"""

    calc_class = calc.CalcSoleilSixsMed2p3v2


class Zaxis(Diffractometer):
    """Z-axis geometry"""

    calc_class = calc.CalcZaxis


class SimMixin(Device):
    """
    Defines `h`, `k`, & `l` pseudo-positioners.

    Use this mixin class with any of the diffractometer geometries to create
    your own simulator.  Follow one of the simulators below, such as
    :class:`~hkl.geometries.SimulatedE4CV`.  You should replace ``E4CV`` with
    your geomtry's name.  And, you will need to create SoftPositioner components
    for each of the real-space axes, in the order required by that geometry.
    """

    h = Cpt(PseudoSingle, "", kind="hinted")
    k = Cpt(PseudoSingle, "", kind="hinted")
    l = Cpt(PseudoSingle, "", kind="hinted")


class SimulatedE4CV(SimMixin, E4CV):
    """SimulatedE4CV: Eulerian 4-circle diffractometer, vertical"""

    omega = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    chi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    phi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    tth = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")


class SimulatedE6C(SimMixin, E6C):
    """SimulatedE6C: Eulerian 6-circle diffractometer"""

    mu = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    omega = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    chi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    phi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    gamma = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    delta = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")


class SimulatedK4CV(SimMixin, K4CV):
    """SimulatedK4CV: Kappa 4-circle diffractometer, vertical"""

    komega = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    kappa = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    kphi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    tth = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")


class SimulatedK6C(SimMixin, K6C):
    """SimulatedK6C: Kappa 6-circle diffractometer"""

    mu = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    komega = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    kappa = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    kphi = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    gamma = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")
    delta = Cpt(SoftPositioner, limits=(-180, 180), init_pos=0, kind="normal")



import numpy.testing
import pytest

import hkl.user
from hkl import SI_LATTICE_PARAMETER
from hkl import SimulatedE4CV

a0 = SI_LATTICE_PARAMETER


class Fourc(SimulatedE4CV):
    """4-circle for testing."""


@pytest.fixture(scope="function")
def fourc():
    fourc = Fourc("", name="fourc")
    fourc.wait_for_connection()
    fourc._update_calc_energy()
    return fourc


def test_select_diffractometer(capsys, fourc):
    # This test function must be first or the next assertion will fail.
    assert hkl.user._geom_ is None
    hkl.user.select_diffractometer(fourc)
    assert hkl.user._geom_ is not None
    assert hkl.user._geom_ == fourc
    capsys.readouterr()  # flush the output buffers

    hkl.user.show_selected_diffractometer()
    out, err = capsys.readouterr()
    assert str(err) == ""
    assert str(out).strip() == "fourc"


def test_cahkl(fourc):
    hkl.user.select_diffractometer(fourc)
    fourc.calc["tth"].limits = (0, 180)

    # use the default "main" sample and UB matrix
    response = hkl.user.cahkl(1, 0, 0)
    expected = (30, 0, 90, 60)
    assert round(response[0]) == expected[0]
    assert round(response[1]) == expected[1]
    assert round(response[2]) == expected[2]
    assert round(response[3]) == expected[3]


def test_cahkl_table(capsys, fourc):
    hkl.user.select_diffractometer(fourc)
    fourc.calc["tth"].limits = (0, 180)

    # use the default "main" sample and UB matrix
    rlist = [(1, 0, 0), (0, 1, 0)]
    hkl.user.cahkl_table(rlist, digits=0)
    out, err = capsys.readouterr()
    expected = """
    ========= ======== ===== === === ===
    (hkl)     solution omega chi phi tth
    ========= ======== ===== === === ===
    (1, 0, 0) 0        30    0   90  60
    (0, 1, 0) 0        30    90  0   60
    ========= ======== ===== === === ===
    """.strip().splitlines()
    assert err == ""
    for el, rl in list(zip(expected[3:5], str(out).strip().splitlines()[3:5])):
        # just compare the position values
        for e, r in list(zip(el.split()[-4:], rl.split()[-4:])):
            assert float(r) == float(e)


def test_calc_UB(fourc):
    hkl.user.select_diffractometer(fourc)
    hkl.user.new_sample("silicon standard", a0, a0, a0, 90, 90, 90)
    r1 = hkl.user.setor(4, 0, 0, tth=69.0966, omega=-145.451, chi=0, phi=0, wavelength=1.54)
    fourc.omega.move(-145.451)
    fourc.chi.move(90)
    fourc.phi.move(0)
    fourc.tth.move(69.0966)
    r2 = hkl.user.setor(0, 4, 0)

    ub = hkl.user.calc_UB(r1, r2)
    if ub is None:
        hkl.user.calc_UB(r1, r2)
        ub = fourc.calc.sample.UB
    assert isinstance(ub, numpy.ndarray)
    assert isinstance(fourc.UB.get(), numpy.ndarray)


def test_list_samples(capsys, fourc):
    hkl.user.select_diffractometer(fourc)
    hkl.user.new_sample("silicon", a0, a0, a0, 90, 90, 90)
    capsys.readouterr()  # flush the output buffers

    hkl.user.list_samples(verbose=False)
    out, err = capsys.readouterr()
    assert err == ""
    expected = """
    silicon (*): [5.431020511, 5.431020511, 5.431020511, 90.0, 90.0, 90.0]
    main: [1.54, 1.54, 1.54, 90.0, 90.0, 90.0]
    """.strip().splitlines()
    for e, r in list(zip(expected, str(out).strip().splitlines())):
        assert r.strip() == e.strip()

    hkl.user.list_samples()
    out, err = capsys.readouterr()
    assert err == ""
    expected = """
    Sample: silicon (*)

    ======= =========================================================
    key     value
    ======= =========================================================
    name    silicon
    lattice [5.431020511, 5.431020511, 5.431020511, 90.0, 90.0, 90.0]
    U       [[1. 0. 0.]
             [0. 1. 0.]
             [0. 0. 1.]]
    UB      [[ 1.15691 -0.      -0.     ]
             [ 0.       1.15691 -0.     ]
             [ 0.       0.       1.15691]]
    ======= =========================================================


    Sample: main

    ======= ====================================
    key     value
    ======= ====================================
    name    main
    lattice [1.54, 1.54, 1.54, 90.0, 90.0, 90.0]
    U       [[1. 0. 0.]
             [0. 1. 0.]
             [0. 0. 1.]]
    UB      [[ 4.07999 -0.      -0.     ]
             [ 0.       4.07999 -0.     ]
             [ 0.       0.       4.07999]]
    ======= ====================================
    """.strip().splitlines()
    for e, r in list(zip(expected, str(out).strip().splitlines())):
        assert r.strip() == e.strip()


def test_new_sample(fourc):
    hkl.user.select_diffractometer(fourc)

    # sample is the silicon standard
    hkl.user.new_sample("silicon standard", a0, a0, a0, 90, 90, 90)
    assert fourc.calc.sample.name == "silicon standard"
    lattice = fourc.calc.sample.lattice
    assert round(lattice.a, 9) == a0
    assert round(lattice.b, 9) == a0
    assert round(lattice.c, 9) == a0
    assert round(lattice.alpha, 7) == 90
    assert round(lattice.beta, 7) == 90
    assert round(lattice.gamma, 7) == 90


def test_set_energy(fourc):
    hkl.user.select_diffractometer(fourc)
    numpy.testing.assert_approx_equal(fourc.energy.get(), 8)
    assert fourc.energy_offset.get() == 0
    assert fourc.energy_units.get() == "keV"
    numpy.testing.assert_approx_equal(fourc.calc.energy, 8)

    hkl.user.set_energy(8.1)
    numpy.testing.assert_approx_equal(fourc.energy.get(), 8.1)
    assert fourc.energy_offset.get() == 0
    assert fourc.energy_units.get() == "keV"
    numpy.testing.assert_approx_equal(fourc.calc.energy, 8.1)

    hkl.user.set_energy(7500, units="eV")
    numpy.testing.assert_approx_equal(fourc.energy.get(), 7500)
    assert fourc.energy_offset.get() == 0
    assert fourc.energy_units.get() == "eV"
    numpy.testing.assert_approx_equal(fourc.calc.energy, 7.5)

    hkl.user.set_energy(7100, units="eV", offset=25)
    numpy.testing.assert_approx_equal(fourc.energy.get(), 7100)
    assert fourc.energy_offset.get() == 25
    assert fourc.energy_units.get() == "eV"
    numpy.testing.assert_approx_equal(fourc.calc.energy, 7.125)

    # Now, do not set offset.  It will use the previous value.
    hkl.user.set_energy(2500, units="eV")
    numpy.testing.assert_approx_equal(fourc.energy.get(), 2500)
    assert fourc.energy_offset.get() == 25
    assert fourc.energy_units.get() == "eV"
    numpy.testing.assert_approx_equal(fourc.calc.energy, 2.525)


def test_setor(fourc):
    hkl.user.select_diffractometer(fourc)
    hkl.user.new_sample("silicon standard", a0, a0, a0, 90, 90, 90)

    assert len(fourc.calc.sample.reflections) == 0
    hkl.user.setor(4, 0, 0, -145.451, 0, 0, 69.0966, wavelength=1.54)
    assert len(fourc.calc.sample.reflections) == 1
    assert fourc.calc.sample.reflections == [(4, 0, 0)]

    fourc.omega.move(-145.451)
    fourc.chi.move(90)
    fourc.phi.move(0)
    fourc.tth.move(69.0966)
    hkl.user.setor(0, 4, 0)
    assert len(fourc.calc.sample.reflections) == 2
    assert fourc.calc.sample.reflections == [(4, 0, 0), (0, 4, 0)]


def test_show_sample(capsys, fourc):
    hkl.user.select_diffractometer(fourc)
    hkl.user.new_sample("silicon", a0, a0, a0, 90, 90, 90)
    capsys.readouterr()  # flush the output buffers

    hkl.user.show_sample(verbose=False)
    out, err = capsys.readouterr()
    assert str(out).strip() == ("silicon (*):" " [5.431020511, 5.431020511, 5.431020511, 90.0, 90.0, 90.0]")
    assert err == ""

    hkl.user.show_sample()
    out, err = capsys.readouterr()
    assert err == ""
    expected = """
    Sample: silicon (*)

    ======= =========================================================
    key     value
    ======= =========================================================
    name    silicon
    lattice [5.431020511, 5.431020511, 5.431020511, 90.0, 90.0, 90.0]
    U       [[1. 0. 0.]
             [0. 1. 0.]
             [0. 0. 1.]]
    UB      [[ 1.15691 -0.      -0.     ]
             [ 0.       1.15691 -0.     ]
             [ 0.       0.       1.15691]]
    ======= =========================================================
    """.strip().splitlines()
    for e, r in list(zip(expected, str(out).strip().splitlines())):
        assert r.strip() == e.strip()


def test_update_sample(capsys, fourc):
    hkl.user.select_diffractometer(fourc)

    hkl.user.update_sample(2, 2, 2, 90, 90, 90)
    out, err = capsys.readouterr()
    assert err == ""
    expected = """
    main (*): [2.0, 2.0, 2.0, 90.0, 90.0, 90.0]
    """.strip().splitlines()
    for e, r in list(zip(expected, str(out).strip().splitlines())):
        assert r.strip() == e.strip()


def test_pa(fourc, capsys):
    hkl.user.select_diffractometer(fourc)

    tbl = hkl.user.pa()
    assert tbl is None
    out, err = capsys.readouterr()
    assert len(out) > 0
    assert err == ""
    out = [v.rstrip() for v in out.strip().splitlines()]
    expected = [
        "===================== ====================================================================",
        "term                  value",
        "===================== ====================================================================",
        "diffractometer        fourc",
        "geometry              E4CV",
        "class                 Fourc",
        "energy (keV)          8.00000",
        "wavelength (angstrom) 1.54980",
        "calc engine           hkl",
        "mode                  bissector",
        "positions             ===== =======",
        "                      name  value",
        "                      ===== =======",
        "                      omega 0.00000",
        "                      chi   0.00000",
        "                      phi   0.00000",
        "                      tth   0.00000",
        "                      ===== =======",
        "constraints           ===== ========= ========== ===== ====",
        "                      axis  low_limit high_limit value fit",
        "                      ===== ========= ========== ===== ====",
        "                      omega -180.0    180.0      0.0   True",
        "                      chi   -180.0    180.0      0.0   True",
        "                      phi   -180.0    180.0      0.0   True",
        "                      tth   -180.0    180.0      0.0   True",
        "                      ===== ========= ========== ===== ====",
        "sample: main          ================ ===================================================",
        "                      term             value",
        "                      ================ ===================================================",
        "                      unit cell edges  a=1.54, b=1.54, c=1.54",
        "                      unit cell angles alpha=90.0, beta=90.0, gamma=90.0",
        "                      [U]              [[1. 0. 0.]",
        "                                        [0. 1. 0.]",
        "                                        [0. 0. 1.]]",
        "                      [UB]             [[ 4.07999046e+00 -2.49827363e-16 -2.49827363e-16]",
        "                                        [ 0.00000000e+00  4.07999046e+00 -2.49827363e-16]",
        "                                        [ 0.00000000e+00  0.00000000e+00  4.07999046e+00]]",
        "                      ================ ===================================================",
        "===================== ====================================================================",
    ]
    assert len(out) == len(expected)
    assert out == expected


def test_wh(fourc, capsys):
    hkl.user.select_diffractometer(fourc)

    tbl = hkl.user.wh()
    assert tbl is None
    out, err = capsys.readouterr()
    assert len(out) > 0
    assert err == ""
    out = [v.rstrip() for v in out.strip().splitlines()]
    expected = [
        "===================== ========= =========",
        "term                  value     axis_type",
        "===================== ========= =========",
        "diffractometer        fourc",
        "sample name           main",
        "energy (keV)          8.00000",
        "wavelength (angstrom) 1.54980",
        "calc engine           hkl",
        "mode                  bissector",
        "h                     0.0       pseudo",
        "k                     0.0       pseudo",
        "l                     0.0       pseudo",
        "omega                 0         real",
        "chi                   0         real",
        "phi                   0         real",
        "tth                   0         real",
        "===================== ========= =========",
    ]
    assert len(out) == len(expected)
    assert out == expected


def test_or_swap(fourc, capsys):
    hkl.user.select_diffractometer(fourc)
    hkl.user.new_sample("silicon standard", a0, a0, a0, 90, 90, 90)
    assert len(fourc.calc.sample.reflections) == 0

    r400 = hkl.user.setor(4, 0, 0, tth=69.0966, omega=-145.451, chi=0, phi=0, wavelength=1.54)
    r040 = hkl.user.setor(0, 4, 0, tth=69.0966, omega=-145.451, chi=0, phi=90, wavelength=1.54)
    assert len(fourc.calc.sample.reflections) == 2
    assert len(fourc.calc.sample._orientation_reflections) == 0

    hkl.user.calc_UB(r400, r040)
    assert len(fourc.calc.sample._orientation_reflections) == 2
    assert fourc.calc.sample._orientation_reflections == [r400, r040]

    hkl.user.or_swap()
    assert fourc.calc.sample._orientation_reflections == [r040, r400]

    fourc.calc.sample.clear_reflections()
    with pytest.raises(ValueError) as exinfo:
        hkl.user.or_swap()
    assert "There are 0 orientation reflection(s)" in str(exinfo.value)

    hkl.user.calc_UB(r400, r040)
    assert len(fourc.calc.sample._orientation_reflections) == 2
